import numpy as np
import time

import parakeet
from parakeet import allpairs, each
from parakeet.testing_helpers import eq, run_local_tests

def python_update_assignments(X, centroids):
  dists = [[np.sum( (x-y) ** 2) for y in centroids] for x in X]
  return np.argmin(dists, 1)

def python_update_centroids(X, assignments, k):
  d = X.shape[1]
  new_centroids = np.zeros((k,d), dtype=X.dtype)
  for i in xrange(k):
    mask = assignments == i
    count = np.sum(mask)
    assigned_data = X[mask]
    if count > 1:
      new_centroids[i, :] = np.mean(assigned_data)
    elif count == 1:
      new_centroids[i,:] = assigned_data[0]

  return new_centroids

def python_kmeans(X, k, maxiters = 100, initial_assignments = None):
  n = X.shape[0]
  if initial_assignments is None:
    assignments = np.random.randint(0, k, size=n)
  else:
    assignments = initial_assignments
  centroids = python_update_centroids(X, assignments, k)
  for _ in xrange(maxiters):
    old_assignments = assignments

    assignments = python_update_assignments(X, centroids)

    if all(old_assignments == assignments):
      break
    centroids = python_update_centroids(X, assignments, k)

  return centroids

def sqr_dist(x,y):
  return sum((x-y) ** 2)

def parakeet_update_assignments(X, centroids):
  dists = allpairs(sqr_dist, X, centroids)

  return np.argmin(dists, 1)

def mean(X):
  return sum(X) / len(X)

def parakeet_update_centroids(X, assignments, k):
  d = X.shape[1]

  new_centroids = np.zeros((k,d), dtype=X.dtype)
  for i in xrange(k):
    mask = (assignments == i)
    count = np.sum(mask)
    assigned_data = X[mask]
    if count == 1:
      new_centroids[i, :] = assigned_data[0]
    elif count > 1:
      new_centroids[i,:] = parakeet.mean(assigned_data)
  return new_centroids

def parakeet_kmeans(X, k, maxiters = 100, initial_assignments = None):
  n = X.shape[0]
  if initial_assignments is None:
    assignments = np.random.randint(0, k, size = n)
  else:
    assignments = initial_assignments

  centroids = python_update_centroids(X, assignments, k)
  for _ in xrange(maxiters):
    old_assignments = assignments
    assignments = parakeet_update_assignments(X, centroids)
    if all(old_assignments == assignments):
      break
    centroids = python_update_centroids(X, assignments, k)

  return centroids

def test_kmeans():
  n = 200
  d = 4
  X = np.random.randn(n*d).reshape(n,d)
  k = 2
  niters = 10
  assignments = np.random.randint(0, k, size = n)
  parakeet_C = parakeet_kmeans(X, k, niters, assignments)
  python_C = python_kmeans(X, k, niters, assignments)
  assert eq(parakeet_C.shape, python_C.shape), \
      "Got back wrong shape, expect %s but got %s" % \
      (python_C.shape, parakeet_C.shape)
  assert eq(parakeet_C, python_C), \
      "Expected %s but got %s" % (python_C, parakeet_C)


if __name__ == '__main__':
  run_local_tests()



import numpy as np

from parakeet import jit 
from parakeet.testing_helpers import run_local_tests, expect

int64_mat = np.reshape(np.arange(6), (2,3)).astype('int64')
int32_mat = int64_mat.astype('int32')

float64_mat = int64_mat.astype('float64')
float32_mat = int64_mat.astype('float32') 
bool_mat = int64_mat % 2

matrices = [int64_mat, int32_mat, float64_mat, float32_mat, bool_mat]

def matmult_comprehensions(X,Y):
  return np.array([[np.dot(x,y) for y in Y.T] for x in X])

def test_matmult_int64_int32():
  expect(matmult_comprehensions, [int64_mat, int32_mat.T], np.dot(int64_mat, int32_mat.T))

def test_matmult_int32_int32():
  expect(matmult_comprehensions, [int32_mat, int32_mat.T], np.dot(int32_mat, int32_mat.T))

def test_matmult_float64_float32():
  expect(matmult_comprehensions, [float64_mat, float32_mat.T], np.dot(float64_mat, float32_mat.T))

def test_matmult_float32_float32():
  expect(matmult_comprehensions, [float32_mat, float32_mat.T], np.dot(float32_mat, float32_mat.T))

def test_matmult_float32_int32():
  expect(matmult_comprehensions, [float32_mat, int32_mat.T], np.dot(float32_mat, int32_mat.T))

def test_matmult_float64_bool():
  expect(matmult_comprehensions, [float64_mat, bool_mat.T], np.dot(float64_mat, bool_mat.T))

def test_matmult_bool_bool():
  expect(matmult_comprehensions, [bool_mat, bool_mat.T], np.dot(bool_mat, bool_mat.T))

def test_matmult_bool_int32():
  expect(matmult_comprehensions, [bool_mat, int32_mat.T], np.dot(bool_mat, int32_mat.T))
      
if __name__ == "__main__":
  run_local_tests()




import numpy as np
from parakeet.testing_helpers import expect, run_local_tests


int_vec = np.array([2,3])
int_mat = np.array([int_vec, int_vec])

bool_vec = np.array([True, True])
bool_mat = np.array([bool_vec, bool_vec])

float32_vec = np.array([1.0, 2.0], dtype='float32')
float32_mat = np.array([float32_vec, float32_vec])

float64_vec = np.array([1.0, 10.0])
float64_mat = np.array([float64_vec, float64_vec])

arrays = [int_vec, bool_vec, float32_vec, float64_mat]

def test_logaddexp2_bool_bool_vec():
  expect(np.logaddexp2, [bool_vec, bool_vec], np.logaddexp2(bool_vec,bool_vec))

def test_logaddexp2_bool_bool_mat():
  expect(np.logaddexp2, [bool_mat, bool_mat], np.logaddexp2(bool_mat,bool_mat))

def test_logaddexp2_int_bool_vec():
  expect(np.logaddexp2, [int_vec, bool_vec], np.logaddexp2(int_vec,bool_vec))

def test_logaddexp2_int_bool_mat():
  expect(np.logaddexp2, [int_mat, bool_mat], np.logaddexp2(int_mat,bool_mat))

def test_logaddexp2_int_int_vec():
  expect(np.logaddexp2, [int_vec, int_vec], np.logaddexp2(int_vec,int_vec))

def test_logaddexp2_int_int_mat():
  expect(np.logaddexp2, [int_mat, int_mat], np.logaddexp2(int_mat,int_mat))

def test_logaddexp2_int_float32_vec():
  expect(np.logaddexp2, [int_vec, float32_vec], np.logaddexp2(int_vec,float32_vec))

def test_logaddexp2_int_float32_mat():
  expect(np.logaddexp2, [int_mat, float32_mat], np.logaddexp2(int_mat,float32_mat))

def test_logaddexp2_int_float64_vec():
  expect(np.logaddexp2, [int_vec, float64_vec], np.logaddexp2(int_vec,float64_vec))

def test_logaddexp2_int_float64_mat():
  expect(np.logaddexp2, [int_mat, float64_mat], np.logaddexp2(int_mat,float64_mat))

def test_logaddexp2_float32_float64_vec():
  expect(np.logaddexp2, [float32_vec, float64_vec], np.logaddexp2(float32_vec,float64_vec))

def test_logaddexp2_float32_float64_mat():
  expect(np.logaddexp2, [float32_mat, float64_mat], np.logaddexp2(float32_mat,float64_mat))

if __name__ == "__main__":
  run_local_tests()




import parakeet
import numpy as np 


def dot(x,y):
    return np.min(x+y)

def matmult_high_level(X,Y):
  return np.array([[dot(x,y) for y in Y.T] for x in X])

def matmult_loops(X,Y,Z):
  m, d = X.shape
  n = Y.shape[1]
  for i in xrange(m):
    for j in xrange(n):
      total = X[i,0] + Y[0,j] 
      for k in xrange(1,d):
        total = min(total, X[i,k] + Y[k,j])
      Z[i,j] = total 
  return Z

n, d = 500, 500
m = 500
X = np.random.randn(m,d)
Y = np.random.randn(d,n)
Z = np.zeros((m,n))
from compare_perf import compare_perf

compare_perf(matmult_high_level, [X,Y], cpython=True, numba=False)
compare_perf(matmult_loops, [X, Y, Z], cpython=False)




from ..builder import build_fn
from ..syntax import  Alloc,  Index, ArrayView, Const, Transpose, Ravel 
from ..syntax.helpers import zero_i64, one_i64, const_int, const_tuple, true, false   
from ..ndtypes import (ScalarT, PtrT, TupleT, ArrayT, ptr_type, Int64)

from transform import Transform


class LowerArrayOperators(Transform):
  """
  Lower first-order array operators such as Ravel, transpose
  into views or copies
  """


  def transform_Array(self, expr):
    """Array literal"""
    array_t = expr.type
    assert isinstance(array_t, ArrayT), "Expected array but got %s" % array_t
    elt_t = array_t.elt_type
    assert isinstance(elt_t, ScalarT), "Expected array to have scalar elements but got %s" % elt_t

    elts = self.transform_expr_list(expr.elts)
    n = len(elts)

    ptr_t = ptr_type(elt_t)
    alloc = Alloc(elt_t, const_int(n), type = ptr_t)
    ptr_var = self.assign_name(alloc, "data")
    array = self.array_view(ptr_var, 
                            shape = const_tuple(n), 
                            strides = const_tuple(1), 
                            offset = zero_i64,
                            nelts = const_int(n))
    for (i, elt) in enumerate(elts):
      self.setidx(array, const_int(i), elt)
    return array 
  
      
  def transform_Len(self, expr):
    return self.shape(expr, 0)

  def transform_AllocArray(self, expr):
    dims = self.transform_expr(expr.shape)
    return self.alloc_array(elt_t = expr.type.elt_type,
                            dims = dims, 
                            name = "array",
                            order = "C", 
                            array_view = True, 
                            explicit_struct = False)
    
  def transform_Reshape(self, expr):
    assert False, "Reshape not implemented"
    
  def transform_Shape(self, expr):

    return self.shape(self.transform_expr(expr.array))
    
  def transform_Strides(self, expr):
    return self.strides(self.transform_expr(expr.array))
  
  """
  Given first-order array constructors, turn them into IndexMaps
  """

  def mk_range_fn(self, start, step, output_type, _range_fn_cache = {}):
    """
    Given expressions for start, stop, and step of an iteration range
    and a desired output type. 
    
    Create a function which maps indices i \in 0...(stop-start)/step into
    f(i) = cast(start + i * step, output_type)
    """
    
      
    start_is_const = start.__class__ is Const
    step_is_const = step.__class__ is Const
    
    key = output_type, (start.type, start_is_const), (step.type, step_is_const)
    if key in _range_fn_cache:
      fn = _range_fn_cache[key]
    else:
      input_types = []
      if not start_is_const: input_types.append(start.type)
      if not step_is_const: input_types.append(step.type)
      input_types.append(Int64)
      fn, builder, input_vars = build_fn(input_types, output_type)
    
      assert len(input_vars) == (3 - start_is_const - step_is_const), \
        "Unexpected # of input vars %d" % len(input_vars)
        
      counter = 0
      if start_is_const:
        inner_start = start 
      else:
        inner_start = input_vars[counter]
        counter += 1
    
      if step_is_const:
        inner_step = step 
      else:
        inner_step = input_vars[counter]
        counter += 1
    
      idx = input_vars[counter]
    
      value = builder.add(builder.mul(idx, inner_step), inner_start)
      builder.return_(builder.cast(value, output_type))
      _range_fn_cache[key] = fn
    
    closure_args = []
    if not start_is_const: 
      closure_args.append(start)
    if not step_is_const: 
      closure_args.append(step)
    return self.closure(fn, closure_args) 
      
  def transform_Range(self, expr):
    start = expr.start 
    stop = expr.stop
    step = expr.step 
    caster = self.mk_range_fn(start, step, expr.type.elt_type)
    nelts = self.elts_in_slice(start, stop, step)
    
    return self.imap(caster, nelts)

  # TODO: Add support for np.ndindex through this operator 
  def transform_NDIndex(self, expr):
    shape = self.transform_expr(expr.shape)
    assert False, "NDIndex not yet implemented"
    
  
  def transform_Ravel(self, expr):
    
    array = expr.array 
    while array.__class__ is Ravel:
      array = expr.array 
    array = self.transform_expr(array)
    strides = self.tuple_elts(self.attr(array, 'strides'))
    shape = self.tuple_elts(self.attr(array, 'shape'))
    any_unit = false
    nelts = one_i64 
    for i in xrange(len(shape)):
      shape_elt = shape[i]
      nelts = self.mul(nelts, shape_elt)
      stride_elt = strides[i]
      any_unit = self.or_(any_unit, self.eq(stride_elt, one_i64))      
    array_result = self.fresh_var(expr.type, prefix = "raveled")
    data = self.attr(array, 'data')
    offset = self.attr(array, 'offset')
    def contiguous(x):
      view =  self.array_view(data, 
                              shape = self.tuple([nelts]), 
                              strides = self.tuple([one_i64]), 
                              offset = offset, 
                              nelts = nelts)
      self.assign(x, view)
    def not_contiguous(x):
      new_array = self.alloc_array(x.type.elt_type, 
                                   dims = self.attr(array, 'shape'),
                                   name = "fresh_array", 
                                   explicit_struct = False, 
                                   array_view = True, 
                                   order = "C")
      self.array_copy(array, new_array)
      data = self.attr(new_array, 'data')
      flat_view = self.array_view(data, 
                              shape = self.tuple([nelts]), 
                              strides = self.tuple([one_i64]), 
                              offset = offset, 
                              nelts = nelts)
      self.assign(x, flat_view)
    self.if_(any_unit, contiguous, not_contiguous, [array_result])
    return array_result 

    
  def transform_Transpose(self, expr):
    if expr.array.__class__ is Transpose:
      return self.transform_expr(expr.array.array)
    
    array = self.transform_expr(expr.array)
    if isinstance(array.type, ScalarT):
      return array 
    assert isinstance(array.type, ArrayT), "Can't transpose %s : %s" % (array, array.type)
    ndims = array.type.rank 
    if ndims <= 1:
      return array 
    assert ndims == 2, "Tranposing arrays of rank %d not yet supported" % (ndims,)
    
    data = self.attr(array, 'data')
    shape = self.shape(array)
    strides = self.strides(array)
    offset = self.attr(array, 'offset')
    stride_elts = self.tuple_elts(strides)
    shape_elts = self.tuple_elts(shape)
    new_shape = self.tuple(tuple(reversed(shape_elts)))
    new_strides = self.tuple(tuple(reversed(stride_elts)))
    size = self.attr(array, 'size')
    return self.array_view(data, new_shape, new_strides, offset, size)
    
  def transform_Where(self, expr):
    assert False, "Where not implemented"
  
  
  def mk_const_fn(self, idx_type, value, _const_fn_cache = {}):
    if isinstance(idx_type, TupleT) and len(idx_type.elt_types) == 1:
      idx_type = idx_type.elt_types[0]
    key = idx_type, value
    if value in _const_fn_cache:
      return _const_fn_cache[key]
    fn, builder, _ = build_fn([idx_type], value.type)
    builder.return_(value)
    _const_fn_cache[key] = fn 
    return fn 
    
  
  def transform_ConstArray(self, expr):
    const_fn = self.mk_const_fn(expr.shape.type, expr.value)
    return self.imap(const_fn, expr.shape)
  
  def transform_ConstArrayLike(self, expr):
    array = self.transform_expr(expr.array)
    shape = self.shape(array)
    const_fn = self.mk_const_fn(shape.type, expr.value)
    return self.imap(const_fn, shape)
  
  
  


import ctypes

#from dsltools import Node

class TypeFailure(Exception):
  def __init__(self, msg):
    self.msg = msg

class IncompatibleTypes(Exception):
  def __init__(self, t1, t2):
    self.t1 = t1
    self.t2 = t2

  def __repr__(self):
    return "IncompatibleTypes(%s, %s)" % (self.t1, self.t2)

  def __str__(self):
    return repr(self)

class Type(object):
  def combine(self, other):
    raise IncompatibleTypes(self, other)

  def children(self):
    return ()
  
  def node_type(self):
    return self.__class__.__name__
  
  def __hash__(self):
    assert False, "Hash function not implemented for type %s" % (self,)

  def __eq__(self, _):
    assert False, "Equality not implemented for type %s" % (self,)
  
  def __ne__(self, other):
    return not (self == other)
  
  def __str__(self):
    return self.__class__.__name__

class AnyT(Type):
  """top of the type lattice, absorbs all types"""

  def __hash__(self):
    return 101
  
  def combine(self, other):
    return self

  def __eq__(self, other):
    return other.__class__ is AnyT

# since there's only one Any type, just create an instance of the same name
Any = AnyT()

class UnknownT(Type):
  """Bottom of the type lattice, absorbed by all other types"""

  _hash = hash("unknown")
  
  def __hash__(self):
    return self._hash 
  
  
  _members = []
  def  combine(self, other):
    return other

  def __eq__(self, other):
    return other.__class__ is UnknownT

#single instance of the Unknown type with same name
Unknown = UnknownT()

def combine_type_list(types):
  common_type = Unknown

  for t in types:
    common_type = common_type.combine(t)
  return common_type



class ConcreteT(Type):
  """
  Type which actually have some corresponding runtime values, as opposed to
  "Any" and "Unknown"
  """

  def from_python(self, py_val):
    return self.ctypes_repr(py_val)

  def to_python(self, internal):
    return internal
  
class ImmutableT(ConcreteT):
  """
  Tuples, closures, None, and scalars have no mutable fields
  """
  pass 
# None is fully described by its type, so the
# runtime representation can just be the number zero

class NoneT(ImmutableT):
  _members = []
  rank = 0
  ctypes_repr = ctypes.c_int64


  def combine(self, other):
    if isinstance(other, NoneT):
      return self
    else:
      raise IncompatibleTypes(self, other)

  def __str__(self):
    return "NoneT"

  def __hash__(self):
    return 0

  def __eq__(self, other):
    return other.__class__ is NoneT
  
  def __repr__(self):
    return str(self)

NoneType = NoneT()



def is_struct(c_repr):
  return type(c_repr) == type(ctypes.Structure)

class FieldNotFound(Exception):
  def __init__(self, struct_t, field_name):
    self.struct_t = struct_t
    self.field_name = field_name
    
  def __str__(self):
    return "FieldNotFound(%s, %s)" % (self.struct_t, self.field_name)

class StructT(Type):
  """All concrete types excluding scalars and pointers"""

  # expect each child class to fill this list
  _fields_ = []

  _repr_cache = {}

  def field_type(self, name):
    for (field_name, field_type) in self._fields_:
      if field_name == name:
        return field_type
    raise FieldNotFound(self, name)

  def field_pos(self, name):
    for (i, (field_name, _)) in enumerate(self._fields_):
      if field_name == name:
        return i
    raise FieldNotFound(self, name)

###################################################
#                                                 #
#             SCALAR NUMERIC TYPES                #
#                                                 #
###################################################

###################################################
# helper functions to implement properties of Python scalar objects
###################################################
def always_zero(x):
  return type(x)(0)

def identity(x):
  return x

class TypeValueT(ImmutableT):
  """
  Materialization of a type into a value 
  """
  rank = 0
  
  def __init__(self, type):
    self.type = type 
    self._hash = hash(self.type)

  
  def __str__(self):
    return "TypeValue(%s)" % self.type 
  
  def __hash__(self):
    return self._hash 
  
  def __eq__(self, other):
    return other.__class__ is TypeValueT and self.type == other.type
  




import numpy as np 
import parakeet 
from parakeet import jit, testing_helpers

@jit
def maxpool(array, pool_size):
  n, w, h = array.shape
  def _inner(args):
    img, x, y = args
    max_val = -1e12
    for i in xrange(0, pool_size):
      for j in xrange(0, pool_size):
        if x + i < w and y + j < h:
          max_val = max(array[img, x + i, y + j])
    return max_val
  return parakeet.imap(_inner, (n, w, h))

def test_maxpool():
  x = np.array([[1,2,3],
                [4,5,6],
                [7,8,9]])
  xs = np.array([x,x])
  expected_single =  np.array([[5, 6, 6],
                               [8, 9, 9],
                               [8, 9, 9]])
  expected = np.array([expected_single, expected_single])
  testing_helpers.expect(maxpool, [xs, 2], expected)

if __name__ == "__main__":
  testing_helpers.run_local_tests()




from parakeet.testing_helpers import expect, run_local_tests 


def is_prime(n):
  for i in xrange(2,n-1):
    if n % i == 0:
      return False 
  return True 

def test_is_prime():
  expect(is_prime, [2], True)  
  expect(is_prime, [3], True)
  expect(is_prime, [4], False)
  expect(is_prime, [5], True)
  expect(is_prime, [6], False) 

if __name__ == '__main__':
    run_local_tests()



from ..analysis import use_count
from ..syntax import Tuple,  Expr
 
from ..ndtypes import (TupleT,  ArrayT, 
                       NoneT, NoneType,  
                       elt_type, ScalarT, 
                       FloatT, 
                       BoolT,  
                       IntT,  Int64, SignedT,
                       PtrT,  
                       ClosureT, 
                       SliceT, ptr_type)
 

import type_mappings
from fn_compiler import FnCompiler
from compile_util import compile_module_from_source
from .. import config as root_config 
import config 

def attr_from_kwargs(obj, kwargs, attr, value = None):
  """
  If an attribute is in the kwargs dictionary, then assign it to the
  given object and remove it from the dictionary, otherwise assign 
  the default value
  """
  if attr in kwargs:
    setattr(obj, attr, kwargs[attr])
    del kwargs[attr]
  else:
    setattr(obj, attr, value)


class PyModuleCompiler(FnCompiler):
  """
  Compile a Parakeet function into a Python module with an 
  entry-point that unboxes all the PyObject inputs, 
  runs a flattened computations, and boxes the result as PyObjects
  """
  def __init__(self, module_entry = True, *args, **kwargs):
    attr_from_kwargs(self, kwargs, 'compiler_cmd')    
    attr_from_kwargs(self, kwargs, 'compiler_flag_prefix')
    attr_from_kwargs(self, kwargs, 'linker_flag_prefix')  
    attr_from_kwargs(self, kwargs, 'src_extension')
    FnCompiler.__init__(self, module_entry = module_entry, *args, **kwargs)
    
  def unbox_scalar(self, x, t, target = None):
    assert isinstance(t, ScalarT), "Expected scalar type, got %s" % t
    if target is None:
      target = "scalar_value"
      
    result = self.fresh_var(t, target)
    if isinstance(t, IntT):
      check = "PyInt_Check"
      if isinstance(t, SignedT):
        get = "PyInt_AsLong"
      else:
        get = "PyInt_AsUnsignedLongMask"
    elif isinstance(t, FloatT):
      check = "PyFloat_Check"
      get = "PyFloat_AsDouble"
    else:
      assert isinstance(t, BoolT), "Unexpected type %s" % t 
      check = "PyBool_Check"
      get = "PyObject_IsTrue"
    
    self.append("""
      if (%(check)s(%(x)s)) { %(result)s = %(get)s(%(x)s); }
      else { PyArray_ScalarAsCtype(%(x)s, &%(result)s); }
    """ % locals())
    return result 
  
  def unbox_array(self, boxed_array, elt_type, ndims, target = "array_value"):
    shape_ptr = self.fresh_var("npy_intp*", "shape_ptr", "PyArray_DIMS((PyArrayObject*)%s)" % boxed_array)
    strides_bytes = self.fresh_var("npy_intp*", "strides_bytes", 
                                   "PyArray_STRIDES( (PyArrayObject*) %s)" % boxed_array)

    #strides_elts = self.fresh_name("strides_elts")
    #self.append("npy_intp %s[%d];" % (strides_elts, ndims))
    bytes_per_elt = elt_type.dtype.itemsize 
    typename = self.array_struct_type(elt_type, ndims)
    result = self.fresh_var(typename, "unboxed_array")
    ptr = self.get_boxed_array_ptr(boxed_array, ptr_type(elt_type))
    self.setfield(result, 'data', ptr)
    for i in xrange(ndims):
      if config.debug:
        self.printf("converting strides %s[%d] = %%ld to %%ld" % (strides_bytes, i), 
                    "%s[%d]" % (strides_bytes, i), "%s[%d] / %d" % (strides_bytes, i, bytes_per_elt))   
      self.setidx("%s.strides" % result, i, "%s[%d] / %d" % (strides_bytes, i, bytes_per_elt))
      self.setidx("%s.shape" % result, i, "%s[%d]" % (shape_ptr, i))
    self.setfield(result, "offset", "0")
    self.setfield(result, "size", "PyArray_Size(%s)" % boxed_array)
    return result 
  
  def unbox_tuple(self, boxed_tuple, tuple_t, target = "tuple_value"):
    if isinstance(tuple_t, TupleT):
      elt_types = tuple(tuple_t.elt_types)
    elif isinstance(tuple_t, ClosureT):
      elt_types = tuple(tuple_t.arg_types)
    else:
      assert False, "Expected tuple type, got %s" % tuple_t
     
    c_struct_t = self.struct_type_from_fields(elt_types)
    unboxed_elts = []
    for i in xrange(len(elt_types)):
      elt_t = elt_types[i]
      elt_typename = self.to_ctype(elt_t)
      elt = self.fresh_var(elt_typename, "unboxed_elt%d" % i, 
                           self.tuple_elt(boxed_tuple, i, elt_types[i], boxed=True))
      unboxed_elts.append(elt)
    elts_str = ", ".join(unboxed_elts)
    return self.fresh_var(c_struct_t, target, "{" + elts_str + "}")
      
  def unbox_slice(self, boxed, t, target = None):
    if target is None: target = "slice"
    typename = self.to_ctype(t)
    start = self.unbox("((PySliceObject*)%s)->start" % boxed, t.start_type, "start")
    stop = self.unbox("((PySliceObject*)%s)->stop" % boxed, t.stop_type, "stop")
    step = self.unbox("((PySliceObject*)%s)->step" % boxed, t.step_type, "step")
    return self.fresh_var(typename, target, "{%s,%s,%s}" % (start,stop,step) )
    
  def unbox(self, boxed, t, target = None):
    if isinstance(t, NoneT):
      return "0"
    elif isinstance(t, PtrT):
      assert False, "Unexpected raw pointer passed as argument from Python %s : %s" % (boxed, t)
    if isinstance(t, ScalarT):
      return self.unbox_scalar(boxed, t, target = target)
    elif isinstance(t, (ClosureT, TupleT)):
      return self.unbox_tuple(boxed, t, target = target)
    elif isinstance(t, SliceT):
      return self.unbox_slice(boxed, t, target = target)
    elif isinstance(t, ArrayT):
      return self.unbox_array(boxed, 
                              elt_type = t.elt_type, 
                              ndims = t.rank, target = target)
    else:
      assert False, "Don't know how to unbox %s : %s" % (boxed, t)
        
  def box_none(self):
    self.append("Py_INCREF(Py_None);")
    return "Py_None"
  
  def box_scalar(self, x, t):  
    if isinstance(t, BoolT):
      return "PyBool_FromLong(%s)" % x
    if x.replace("_", "").isalpha():
      scalar = x
    else:
      scalar = self.fresh_name("scalar");
      self.append("%s %s = %s;" % (self.to_ctype(t), scalar, x))
    return "PyArray_Scalar(&%s, PyArray_DescrFromType(%s), NULL)" % (scalar, type_mappings.to_dtype(t) )
  
  def box_tuple(self, x, t):
    if isinstance(t, ClosureT):
      elt_types = t.arg_types 
    else:
      assert isinstance(t, TupleT)
      elt_types = t.elt_types
    unboxed_elts = self.tuple_elts(x, elt_types)
    boxed_elts = [self.box(elt, elt_t) for elt, elt_t in zip(unboxed_elts, elt_types)]
    n = len(boxed_elts)
    if n == 0:
      return "PyTuple_Pack(0)"
    else:
      return "PyTuple_Pack(%d, %s)" % (n, ", ".join(boxed_elts))
  
  def box_slice(self, x, t):
    start = self.box("%s.start" % x, t.start_type)
    stop = self.box("%s.stop" % x, t.stop_type)
    step = self.box("%s.step" % x, t.step_type)
    return "PySlice_New(%s, %s, %s)" % (start, stop, step)
  
  def box(self, x, t):
    if isinstance(t, ScalarT):
      return self.box_scalar(x, t)
    elif isinstance(t, NoneT):
      return self.box_none()
    elif isinstance(t, (ClosureT, TupleT)):
      return self.box_tuple(x, t)
    elif isinstance(t, SliceT):
      return self.box_sclie(x, t)
    elif isinstance(t, ArrayT):
      return self.box_array(x, t)
    else:
      # all other types already boxed 
      return x
      
      
  def as_pyobj(self, expr):
    """
    Compile the expression and if necessary box it up as a PyObject
    """
    x = self.visit_expr(expr)
    if isinstance(expr.type, (NoneT, ScalarT, TupleT, ClosureT, ArrayT, SliceT)):
      return self.box(x, expr.type)
    else:
      return x
  
  def as_pyobj_list(self, exprs):
    return [self.as_pyobj(expr) for expr in exprs]
  
  def c_str(self, obj):
    return "PyString_AsString(PyObject_Str(%s))" % obj
  
  def c_type_str(self, obj):
    return self.c_str("PyObject_Type((PyObject*) %s)" % obj)
   
  def print_pyobj(self, obj, text = ""):
    text = '"%s"' % text
    self.append('printf("%%s%%s\\n", %s, %s);' % (text, self.c_str(obj)))
  
  def print_pyobj_type(self, obj, text=""):
    text = '"%s"' % text
    self.append('printf("%%s%%s\\n", %s, %s);' % (text, self.c_type_str(obj)))

    
  def setfield(self, base, fieldname, value):
    self.append("%s.%s = %s;" % (base, fieldname, value))
  
  def getidx(self, arr, idx, full_array = False):
    """
    Given either the compiled string representation or Parakeet IR
    for an array and index, construct a C indexing expression 
    """

    if isinstance(arr, Expr):
      if isinstance(arr.type, ArrayT):
        arr = self.visit_expr(arr)
        ptr = "%s.data.raw_ptr" % arr
      else:
        assert isinstance(arr.type, PtrT), \
          "Expected array or pointer but got %s : %s" % (arr, arr.type)
        arr = None
        ptr = self.visit_expr(arr)
    else:    
      assert isinstance(arr, str), "Expected string repr of array but got %s" % arr
      if full_array:
        ptr = "%s.data.raw_ptr" % arr 
      else:
        ptr = arr 
        arr = None  
        
    if isinstance(idx, Expr):
      if isinstance(idx.type, TupleT):
        nelts = len(idx.type.elt_types)
        tuple_value = self.visit_expr(idx)
        idx = ["%s.elt%d" % (tuple_value,i) for i in xrange(nelts)]
      else:
        assert isinstance(idx.type, ScalarT), "Expected index %s to be scalar or tuple" % idx
        offset = idx 
        
    if isinstance(idx, str):
      offset = idx 
    elif isinstance(idx, (int,long)):
      offset = "%s" % idx
    elif isinstance(idx, (list,tuple)):
      assert arr, "Expected full array but only pointer %s is available" % ptr 
      offset = "0"
      for i,idx_elt in enumerate(idx):
        if isinstance(idx_elt, Expr):
          idx_elt = self.visit_expr(idx_elt)
        offset = self.add(offset, self.mul("%s.strides[%d]" % (arr,i), idx_elt))
      
    return "%s[%s]" % (ptr, offset)
  
  def setidx(self, arr, idx, value, full_array = False, return_stmt = False):
    stmt = "%s = %s;" % (self.getidx(arr, idx, full_array = full_array), value)
    if return_stmt:
      return stmt 
    else:
      self.append(stmt)

     
  def tuple_to_stack_array(self, expr, name = "array_from_tuple", elt_type = None):
    t0 = expr.type.elt_types[0]
    
    assert expr.type.__class__ is TupleT 
    assert all(t == t0 for t in expr.type.elt_types[1:])
    
    if expr.__class__ is Tuple:
      elts = [self.visit_expr(elt_expr) for elt_expr in expr.elts]
    else:
      tup = self.visit_expr(expr)
      self.check_tuple(tup)
      elts = self.tuple_elts(tup, expr.type.elt_types)
    
    array_name = self.fresh_name(name)
    n = len(expr.type.elt_types)
    if elt_type is None:
      elt_type = self.to_ctype(t0)
    self.append("%s %s[%d];" % (elt_type, array_name, n))
    for i, elt in enumerate(elts):
      self.append("%s[%d] = %s;" % (array_name, i, elt))
    return array_name
    
  def array_to_tuple(self, arr, n, elt_t, boxed = False):
    raw_elts = ["%s[%d]" % (arr,i) for i in xrange(n)]
    if boxed:
      if n == 0: 
        return "PyTuple_Pack(0);"
      else:
        boxed_elts = [self.box_scalar(raw_elt, elt_t) for raw_elt in raw_elts]
        elt_str = ", ".join(boxed_elts)
        return "PyTuple_Pack(%d, %s)" % (n, elt_str)
    else:
      field_types = (self.to_ctype(elt_t),) * n 
      tuple_struct_t = self.struct_type_from_fields(field_types)
      init = "{%s}" % ", ".join(raw_elts)
      return self.fresh_var(tuple_struct_t, "tuple_value", init)
      
  
  def tuple_elts(self, tup, ts, boxed = False):
    result = []
    for i,t in enumerate(ts):
      result.append(self.tuple_elt(tup, i, t, boxed = boxed))
    return result
  
  def mk_tuple(self, elts, boxed = False):
    n = len(elts)
    if boxed:
      if n == 0: 
        return "PyTuple_Pack(0);"
      else:
        elt_str = ", ".join(self.as_pyobj_list(elts)) 
        return "PyTuple_Pack(%d, %s)" % (n, elt_str)
    else:
      field_types = tuple(elt.type for elt in elts)
      tuple_struct_t = self.struct_type_from_fields(field_types)
      elts_str = ", ".join(self.visit_expr(elt) for elt in elts)
      init = "{" + elts_str +  "}"
      return self.fresh_var(tuple_struct_t, "tuple_value", init)
    
  
  def check_tuple(self, tup):
    if not config.check_pyobj_types: return 
    self.newline()
    self.comment("Checking tuple type for %s" % tup)
    self.append("""
      if (!PyTuple_Check(%s)) { 
        PyErr_Format(PyExc_AssertionError, 
                    "Expected %s to be tuple, got %%s", 
                    %s); 
        return NULL;
      }""" % (tup, tup, self.c_type_str(tup)))
 
  def check_slice(self, obj):
    if not config.check_pyobj_types: return 
    self.newline()
    self.comment("Checking slice type for %s" % obj)
    self.append("""
      if (!PySlice_Check(%s)) { 
        PyErr_Format(PyExc_AssertionError, 
                    "Expected %s to be slice, got %%s", 
                    %s); 
        return NULL;
      }""" % (obj, obj, self.c_type_str(obj)))
  
  def check_array(self, arr):
    if not config.check_pyobj_types: return 
    self.newline()
    self.comment("Checking array type for %s" % arr)
    self.append("""
      if (!PyArray_Check(%s)) { 
        PyErr_Format(PyExc_AssertionError, 
                    "Expected %s to be array, got %%s : %%s", 
                    %s, %s); 
        return NULL;
      }""" % (arr, arr, self.c_str(arr), self.c_type_str(arr)))
  
  
  def check_bool(self, x):
    if not config.check_pyobj_types: return 
    self.newline()
    self.comment("Checking bool type for %s" % x)
    self.append("""
      if (!PyArray_IsScalar(%s, Bool)) { 
        PyErr_Format(PyExc_AssertionError, 
                     "Expected %s to be bool, got %%s", 
                     %s); 
        return NULL;
      }""" % (x, x, self.c_type_str(x)))
  
  def check_int(self, x):
    if not config.check_pyobj_types: return 
    self.newline()
    self.comment("Checking int type for %s" % x)
    self.append("""
      if (!PyArray_IsIntegerScalar(%s)) { 
        PyErr_Format(PyExc_AssertionError, 
                     "Expected %s to be int, got %%s", 
                     %s); 
        return NULL;
      }""" % (x, x, self.c_type_str(x)))
  
  def check_type(self, v, t):
    if not config.check_pyobj_types: return 
    if isinstance(t, (ClosureT, TupleT)):
      self.check_tuple(v)
    elif isinstance(t, BoolT):
      self.check_bool(v)
    elif isinstance(t, IntT):
      self.check_int(v)
    elif isinstance(t, ArrayT):
      self.check_array(v)
      
  def tuple_elt(self, tup, idx, t, boxed = False):
    if boxed: 
      self.check_tuple(tup)
      proj_str = "PyTuple_GetItem(%s, %d)" % (tup, idx)
      elt_obj = self.fresh_var("PyObject*", "%s_elt" % tup, proj_str)
      result = self.unbox(elt_obj, t)
      if config.debug and t == Int64:
        self.append(""" printf("tupleproj %s[%d] = %%" PRId64 "\\n", %s);""" % (tup, idx, result))
      return result
    else:
      return "%s.elt%d" % (tup, idx)  
 
    
  def decref(self, obj):
    self.append("Py_DECREF(%s);" % obj)
  
  def get_boxed_array_ptr(self, v, parakeet_ptr_t):
    self.check_array(v)
    c_struct_type = self.to_ctype(parakeet_ptr_t)
    c_ptr_type = type_mappings.to_ctype(parakeet_ptr_t) 
    data_field = "(%s) PyArray_DATA(((PyArrayObject*) %s))" % (c_ptr_type, v) 
    # get the data field but also fill the base object 
    return self.fresh_var(c_struct_type, "data", "{%s, %s}" % (data_field, v))
 
  def attribute(self, v, attr, t, boxed = False):
    if attr == "data":
      if boxed:
        self.check_array(v)
        struct_type = self.to_ctype(t)
        ptr_type = type_mappings.to_ctype(t) 
        data_field = "(%s) PyArray_DATA(((PyArrayObject*) %s))" % (ptr_type, v) 
        # get the data field but also fill the base object 
        return self.fresh_var(struct_type, "data", "{%s, %s}" % (data_field, v))
      else:
        return "%s.data" % v  
    elif attr == "shape":
      shape_name = self.fresh_name("shape")
      elt_types = t.elt_types
      n = len(elt_types)
      elt_t = elt_types[0]  
      if boxed:
        self.check_array(v)
        
        assert all(t == elt_t for t in elt_types)

        shape_array = "PyArray_DIMS( (PyArrayObject*) %s)" % v
        self.append("npy_intp* %s = %s;" % (shape_name, shape_array))
      else:
        self.append("npy_intp* %s = %s.shape;" % (shape_name, v))
      return self.array_to_tuple(shape_name, n, elt_t)
      
    elif attr == "strides":
      elt_types = t.elt_types
      n = len(elt_types)  
      elt_t = elt_types[0]
      if boxed:
        assert False, "Can't directly use NumPy strides without dividing by itemsize"
      else:
        strides_name = self.fresh_var("npy_intp*", "strides", "%s.strides" % v)
      return self.array_to_tuple(strides_name, n, elt_t)
    elif attr == 'offset':
      if boxed:
        return "0"
      else:
        return "%s.offset" % v
    elif attr in ('size', 'nelts'):
      if boxed:
        return "PyArray_Size(%s)" % v
      else:
        return "%s.size" % v
       
    elif attr in ('start', 'stop', 'step'):
      if boxed:
        self.check_slice(v)
        obj = "((PySliceObject*)%s)->%s" % (v, attr)
        return self.unbox_scalar(obj, t, attr)
      else:
        return "%s.%s" % (v, attr)
    else:
      assert False, "Unsupported attribute %s" % attr 
   
    
  
  def alloc_array(self, array_t, shape_expr):
    if isinstance(shape_expr.type, ScalarT):
      dim = self.visit_expr(shape_expr)
      nelts = dim 
      shape_elts = [dim]
      ndims = 1 
    else:
      shape = self.visit_expr(shape_expr)
      ndims = array_t.rank
      shape_elts = ["%s.elt%d" % (shape, i) for i in xrange(ndims)]
      nelts = self.fresh_var("int64_t", "nelts", " * ".join(shape_elts))
    bytes_per_elt = array_t.elt_type.dtype.itemsize
    typename = self.to_ctype(array_t)
    result = self.fresh_var(typename, "new_array")
    raw_ptr_t = self.to_ctype(array_t.elt_type) + "*"
    self.setfield(result, "data.raw_ptr", "(%s) malloc(%s * %s)" % (raw_ptr_t, nelts, bytes_per_elt) )
    self.setfield(result, "data.base", "(PyObject*) NULL")
    self.setfield(result, "offset", "0")
    self.setfield(result, "size", nelts)
    # assume C-order layout
    strides_elts = [" * ".join(["1"] + shape_elts[(i+1):]) for i in xrange(ndims)]
    for i in xrange(ndims):
      self.setidx("%s.shape" % result, i, shape_elts[i])
      # assume c major layout for now
      self.setidx("%s.strides" % result, i, strides_elts[i])
    if config.debug:
      self.printf("[Debug] Done allocating array")
    return result 
    
  
  def visit_AllocArray(self, expr, boxed=False):
    if boxed:
      shape = self.tuple_to_stack_array(expr.shape)
      t = type_mappings.to_dtype(elt_type(expr.type))
      return "(PyArrayObject*) PyArray_SimpleNew(%d, %s, %s)" % (expr.type.rank, shape, t)
    
    if config.debug:
      print "[Debug] Allocating array : %s " % expr.type  
    return self.alloc_array(expr.type, expr.shape)
     
  def visit_Tuple(self, expr):
    return self.mk_tuple(expr.elts, boxed = False)
  
  def visit_Closure(self, expr):
    return self.mk_tuple(expr.args)
  
  def visit_TupleProj(self, expr):
    tup = self.visit_expr(expr.tuple)
    result = self.tuple_elt(tup, expr.index, expr.type)
    return result
  
  def visit_ClosureElt(self, expr):
    clos = self.visit_expr(expr.closure)
    return self.tuple_elt(clos, expr.index, expr.type)
  
  def visit_ArrayView(self, expr):
    typename = self.to_ctype(expr.type)
    data = self.visit_expr(expr.data)
    ndims = expr.type.rank 
    offset = self.visit_expr(expr.offset)
    count = self.visit_expr(expr.size)
    shape = self.visit_expr(expr.shape)
    strides = self.visit_expr(expr.strides)
  
    strides_elts = ["%s.elt%d" % (strides, i) for i in xrange(ndims)]
    shape_elts = ["%s.elt%d" % (shape, i) for i in xrange(ndims)]

    result = self.fresh_var(typename, "array_result")
    self.setfield(result, "data", data)
    self.setfield(result, "offset", offset)
    self.setfield(result, 'size', count)
    for i in xrange(ndims):
      self.setidx("%s.shape" % result, i, shape_elts[i])
      self.setidx("%s.strides" % result, i, strides_elts[i])
    return result 
  
  """
  def box_ArrayView(self, expr):
    data = self.visit_expr(expr.data)
    ndims = expr.type.rank 
    offset = self.visit_expr(expr.offset)
    count = self.visit_expr(expr.size)
    shape = self.visit_expr(expr.shape)
    strides = self.visit_expr(expr.strides)
  
    strides_elts = ["%s.elt%d" % (strides, i) for i in xrange(ndims)]
    shape_elts = ["%s.elt%d" % (shape, i) for i in xrange(ndims)]
    shape_array = None 
    strides_array = None 
    elt_type = expr.type.elt_type 
    return self.make_boxed_array(elt_type, ndims, data, strides_array, shape_array, offset, count)
  """
  
  def box_array(self, arr, t):
    elt_t = t.elt_type
    ndims = t.rank 
    data_ptr = "%s.data.raw_ptr" % arr 
    base = "%s.data.base" % arr
    strides_array = "%s.strides" % arr 
    shape_array = "%s.shape" % arr 
    offset = "%s.offset" % arr 
    size = "%s.size" % arr 
    return self.make_boxed_array(elt_t, ndims, data_ptr, strides_array, shape_array, offset, size, base)
  
  def make_boxed_array(self, elt_type, ndims, data_ptr, strides_array, shape_array, offset, size, base = "NULL"):
    
    
    # PyObject* PyArray_SimpleNewFromData(int nd, npy_intp* dims, int typenum, void* data)
    typenum = type_mappings.to_dtype(elt_type)
    array_alloc = \
      "(PyArrayObject*) PyArray_SimpleNewFromData(%d, %s, %s, &%s[%s])" % \
        (ndims, shape_array, typenum, data_ptr, offset)
    vec = self.fresh_var("PyArrayObject*", "fresh_array", array_alloc) 
    self.return_if_null(vec)
    
    # if the pointer had a PyObject reference, 
    # set that as the new array's base 
    if base not in ("0", "NULL"):
      self.append("""
        if (%s) { 
          PyArray_SetBaseObject(%s, %s); 
          Py_INCREF(%s);  
          PyArray_CLEARFLAGS(%s, NPY_ARRAY_OWNDATA); 
        }""" % (base, vec, base, base, vec))
        
    numpy_strides = self.fresh_var("npy_intp*", "numpy_strides")
    self.append("%s = PyArray_STRIDES(  (PyArrayObject*) %s);" % (numpy_strides, vec))
    bytes_per_elt = elt_type.dtype.itemsize
    
    for i in xrange(ndims):
      self.append("%s[%d] = %s[%d] * %d;" % (numpy_strides, i, strides_array, i, bytes_per_elt) )
      
    
    self.append("""
      // clear both fortran and c layout flags 
      PyArray_CLEARFLAGS((PyArrayObject*) %(vec)s, NPY_ARRAY_F_CONTIGUOUS);  
      PyArray_CLEARFLAGS((PyArrayObject*) %(vec)s, NPY_ARRAY_C_CONTIGUOUS);
    """ % locals())
    
    f_layout_strides = ["1"]
    for i in xrange(1, ndims):
      shape_elt = "%s[%d]" % (shape_array, i)
      f_layout_strides.append(f_layout_strides[-1] + " * " + shape_elt)
    
    c_layout_strides = ["1"]
    for i in xrange(ndims-1,0,-1):
      shape_elt = "%s[%d]" % (shape_array, i)
      c_layout_strides = [c_layout_strides[-1] + " * " + shape_elt] + c_layout_strides
    
    
    strides_elts = ["%s[%d]" % (strides_array, i) for i in xrange(ndims)]
    is_c_layout = "&& ".join(self.eq(actual, ideal, Int64) 
                             for actual, ideal 
                             in zip(strides_elts, c_layout_strides))
    is_f_layout = " && ".join(self.eq(actual, ideal, Int64) 
                             for actual, ideal 
                             in zip(strides_elts, f_layout_strides))
    
       
    # make sure the contiguity flags are set correctly 
    self.append("""
      // it's possible that *neither* of the above flags should be on
      // which is why we enable them separately here 
      if (%(is_f_layout)s) { PyArray_ENABLEFLAGS((PyArrayObject*)%(vec)s, NPY_ARRAY_F_CONTIGUOUS); }
      if (%(is_c_layout)s) { PyArray_ENABLEFLAGS((PyArrayObject*)%(vec)s, NPY_ARRAY_C_CONTIGUOUS); }
    """ % locals())
    return vec
  
    
  def visit_Attribute(self, expr):
    attr = expr.name
    v = self.visit_expr(expr.value) 
    return self.attribute(v, attr, expr.type)
  
  def visit_Return(self, stmt):
    if self.module_entry:
      v = self.as_pyobj(stmt.value)
      if config.debug: 
        self.print_pyobj_type(v, "Return type: ")
        self.print_pyobj(v, "Return value: ")
      return "return (PyObject*) %s;" % v
    else:
      return FnCompiler.visit_Return(self, stmt)
  
  def visit_block(self, stmts, push = True):
    if push: self.push()
    for stmt in stmts:
      s = self.visit_stmt(stmt)
      self.append(s)
    self.append("\n")
    return self.pop()
  
  
  def enter_module_body(self):
    """
    Some derived compiler classes might want to use this hook
    to generate special code when entering the function body of the module entry
    """
    pass 
  
  def exit_module_body(self):
    pass 
  
  def visit_fn(self, fn):
    if config.print_input_ir:
      print "=== Compiling to C with %s (entry function) ===" % self.__class__.__name__ 
      print fn
    c_fn_name = self.fresh_name(fn.name)
    uses = use_count(fn)
    self.push()
    
    
    dummy = self.fresh_name("dummy")
    args = self.fresh_name("args")
    
    if config.debug: 
      self.newline()
      self.printf("\\nStarting %s : %s..." % (c_fn_name, fn.type))
      
    for i, argname in enumerate(fn.arg_names):
      assert argname in uses, "Couldn't find arg %s in use-counts" % argname
      if uses[argname] <= 1:
        self.comment("Skipping unused argument %s" % argname)
        continue
      self.comment("Unpacking argument %s"  % argname)
      c_name = self.name(argname)
      t = fn.type_env[argname]
      self.comment("Getting arg #%d: %s : %s => %s" % (i+1, argname,  t, c_name) )
      self.append("PyObject* %s = PyTuple_GetItem(%s, %d);" % (c_name, args, i))
      
      self.check_type(c_name, t)
      if config.debug:
        self.printf("Printing arg #%d %s" % (i,c_name))
        self.print_pyobj_type(c_name, text = "Type: ")
        self.print_pyobj(c_name, text = "Value: ")
      
      if isinstance(t, (TupleT, NoneT, PtrT, ClosureT, ScalarT, ArrayT, SliceT)):
        #new_name = self.name(argname, overwrite = True)
        self.comment("Unboxing %s : %s" % (argname, t))
        var = self.unbox(c_name, t, target = argname)

        self.name_mappings[argname] = var
      

    self.enter_module_body()
    c_body = self.visit_block(fn.body, push=False)
    self.exit_module_body()
    c_body = self.indent(c_body )
    c_args = "PyObject* %s, PyObject* %s" % (dummy, args) #", ".join("PyObject* %s" % self.name(n) for n in fn.arg_names)
    c_sig = "PyObject* %(c_fn_name)s (%(c_args)s)" % locals() 
    fndef = "%s {\n\n %s}" % (c_sig, c_body)
    return c_fn_name, c_sig, fndef 
  
  _entry_compile_cache = {} 
  def compile_entry(self, parakeet_fn):  
    # we include the compiler's class as part of the key
    # since this function might get reused by descendant backends like OpenMP and CUDA
    key = parakeet_fn.cache_key, self.__class__ 
    compiled_fn = self._entry_compile_cache.get(key)
    if compiled_fn: return compiled_fn 
    
    name, sig, src = self.visit_fn(parakeet_fn)
    
    if config.print_function_source: 
      print "Generated C source for %s: %s" %(name, src)
    ordered_function_sources = [self.extra_functions[extra_sig] for 
                                extra_sig in self.extra_function_signatures]

    compiled_fn = compile_module_from_source(
      src, 
      fn_name = name,
      fn_signature = sig, 
      src_extension = self.src_extension,
      extra_objects = set(self.extra_objects),
      extra_function_sources = ordered_function_sources, 
      declarations =  self.declarations, 
      extra_compile_flags = self.extra_compile_flags, 
      extra_link_flags = self.extra_link_flags, 
      print_source = root_config.print_generated_code, 
      compiler = self.compiler_cmd, 
      compiler_flag_prefix = self.compiler_flag_prefix, 
      linker_flag_prefix = self.linker_flag_prefix)
    self._entry_compile_cache[key]  = compiled_fn
    return compiled_fn





from .. import syntax, prims 
from ..analysis import OffsetAnalysis, SyntaxVisitor
from ..ndtypes import  ArrayT, ScalarT, SliceT, TupleT, NoneT
from ..syntax import unwrap_constant, Expr  

import shape
import shape_from_type
from shape import (Var, Const, Shape, Tuple, Closure, Slice, Scalar, Unknown, 
                   ConstSlice, Struct, AnyScalar, Add, Mult, Div, Sub, Mod, 
                   any_scalar, unknown_value, const, any_value, combine_list, 
                   increase_rank, make_shape, is_zero, is_one, Ptr,
                   dims, combine_dims     
                   ) 

class ShapeInferenceFailure(Exception):
  def __init__(self, value, fn):
    self.value = value 
    self.fn = fn 
  
  def __str__(self):
    return "Couldn't infer shape of %s in function %s" % (self.value, self.fn.name)

  
counter = 0
class ShapeInference(SyntaxVisitor):
  def size_along_axis(self, value, axis):
    assert isinstance(value, Shape)
    return value.dims[axis]

  
  def is_tuple(self, x):
    return isinstance(x, Tuple)

  def is_none(self, x):
    return isinstance(x, Const) and x.value is None

  def rank(self, value):
    if isinstance(value, Shape):
      return value.rank
    else:
      return 0
  
  def max_rank(self, values):
    return max(self.rank(v) for v in values)
  
  def int(self, x):
    return const(x)

  def bool(self, x):
    return const(x)


  _scalar_shape_classes = (Const, Var, Add, Sub, Mult, Div, Mod,  AnyScalar)
  def add(self, x, y):
    cx = x.__class__ 
    cy = y.__class__ 

    if cx in self._scalar_shape_classes and cy in self._scalar_shape_classes:
      if is_zero(x):
        return y
      elif is_zero(y):
        return x
      elif cx is Const and cy is Const: 
        return const(x.value + y.value)
      elif cx is AnyScalar or cy is AnyScalar:
        return any_scalar
      else:
        return Add(x,y)
    else:
      return any_value 

  def sub(self, x, y):
    cx = x.__class__ 
    cy = y.__class__ 
    if cx in self._scalar_shape_classes and cy in self._scalar_shape_classes:
      if is_zero(y):
        return x
      elif cx is Const and cy is Const: 
        return const(x.value - y.value)
      elif cx is AnyScalar or cy is AnyScalar:
        return any_scalar
      elif x == y: 
        return Const(0)
      else:
        return Sub(x, y)
    else:
      return any_value 

  def mul(self, x, y):
    cx = x.__class__ 
    cy = y.__class__ 

    if cx in self._scalar_shape_classes and cy in self._scalar_shape_classes:
      if is_zero(x) or is_zero(y):
        return const(0)
      elif is_one(x):
        return y
      elif is_one(y):
        return x
      elif cx is AnyScalar or cy is AnyScalar:
        return any_scalar
      else:
        return Mult(x,y)
    else:
      return any_value
    
  def div(self, x, y):
    assert not is_zero(y), "Encountered divide by zero during shape inference"
    cx = x.__class__ 
    cy = y.__class__ 

    if cx in self._scalar_shape_classes and cy in self._scalar_shape_classes:
      if is_one(y):
        return x
      elif cx is AnyScalar or cy is AnyScalar:
        return any_scalar
      elif cx is Const and cy is Const:
        return const(int(x.value / y.value))
      elif x == y:
        return const(1)
      else:
        return Div(x, y)
    else:
      return any_value 
      

  def shape(self, x):
    if isinstance(x, Shape):
      return Tuple(x.dims)
    else:
      return Tuple(())

  def elt_type(self, x):
    return "DON'T CARE ABOUT ELT TYPES"

  def alloc_array(self, _, dims):
    return make_shape(dims)

  def index(self, arr, idx):

    if isinstance(arr, Scalar):
      return arr
    assert arr.__class__ is Shape
    if isinstance(idx, (Scalar, Slice, ConstSlice)):
      indices = [idx]
    elif idx.__class__ is Tuple:
      indices = idx.elts
    else:
      assert False, "Unexpected index: %s" % (idx,)
    result_dims = []
    for (i, curr_idx) in enumerate(indices):
      old_dim = arr.dims[i]
      if curr_idx is None or \
         (isinstance(curr_idx, Const) and curr_idx.value is None):
        result_dims.append(old_dim)
      elif isinstance(curr_idx, Scalar):
        pass
      elif curr_idx.__class__ is ConstSlice:
        result_dims.append(curr_idx.nelts)
      elif curr_idx.__class__ is Shape:
        if len(curr_idx.dims) == 0:
          # same as unknown scalar 
          pass 
        else:
          assert len(curr_idx.dims) == 1, "Indexing by a multi-dimensional array not yet supported"
          result_dims.append(curr_idx.dims[0])
      else:
        assert curr_idx.__class__ is Slice, "Unsupported index %s" % curr_idx

        if curr_idx.start is None:
          lower = const(0)
        elif isinstance(curr_idx.start, Const):
          if curr_idx.start.value is None:
            lower = const(0)
          elif curr_idx.start.value < 0:
            lower = self.sub(old_dim, curr_idx.start)
          else:
            lower = curr_idx.start
        else:
          lower = any_scalar
         
        if curr_idx.stop is None:
          upper = old_dim 
        elif isinstance(curr_idx.stop, Const):
          if curr_idx.stop.value is None:
            upper = old_dim
          elif curr_idx.stop.value < 0:
            upper = self.sub(old_dim, curr_idx.stop)
          else:
            upper = curr_idx.stop
        else:
          upper = any_scalar

        n = self.sub(upper, lower)
        step = curr_idx.step
        if step and \
            isinstance(step, Const) and \
            step.value is not None and \
            step.value != 1:
          n = self.div(n, step)
        result_dims.append(n)
    n_original = len(arr.dims)
    n_idx= len(indices)
    if n_original > n_idx:
      result_dims.extend(arr.dims[n_idx:])

    return make_shape(result_dims)
  
  def slice_along_axis(self, arr, axis):
    if arr.__class__ is Shape:
      dims = arr.dims[:axis] + arr.dims[(axis+1):]
      if len(dims) > 0:
        return Shape(dims)
      else:
        return any_scalar
    else:
      return arr 
  
  def tuple(self, elts):
    return Tuple(tuple(elts))

  def concat_tuples(self, t1, t2):
    return Tuple(t1.elts + t2.elts)

  def setidx(self, arr, idx, v):
    pass

  def loop(self, start_idx, stop_idx, body):
    body(start_idx)

  class Accumulator(object):
    def __init__(self, v):
      self.v = v

    def update(self, new_v):
      self.v = new_v

    def get(self):
      return self.v

  def accumulate_loop(self, start_idx, stop_idx, body, init):
    acc = self.Accumulator(init)
    body(acc, start_idx)
    return acc.get()

  def check_equal_sizes(self, sizes):
    pass

  def slice_value(self, start, stop, step):
    
    # if all elements of the slice are constant 
    # then we can ignore the exact start/stop/step 
    # and track only the number of elements in the 
    # slice
    if start.__class__ is Const and \
       stop.__class__ is Const and \
       stop.value is not None and \
       step.__class__ is Const:
      start_val = start.value
      if start_val is None:
        start_val = 0
      step_val = step.value
      if step_val is None:
        step_val = 1
      nelts = (stop.value - start_val) / step_val
      # TODO: 
      # Properly handle negative slicing 
      if nelts >= 0:
        return ConstSlice(nelts)
    return Slice(start, stop, step)

  def call(self, fn, args):
    if fn.__class__ is Closure:
      args = tuple(fn.args) + tuple(args)
      fn = fn.fn
    return symbolic_call(fn, args)

  def invoke(self, fn, args):
    return self.call(fn, args)
  none = None
  null_slice = slice(None, None, None)

  def identity_function(self, x):
    return x
  
  def visit_fn(self, fn):
    assert isinstance(fn, syntax.TypedFn), "Expected typed function, got %s" % fn 
    self.fn = fn 
    self.value_env = {}
    self.equivalence_classes = {}

    self.known_offsets = OffsetAnalysis().visit_fn(fn)

    arg_types = [fn.type_env[name] for name in fn.arg_names]
    input_values = shape_from_type.Converter().from_types(arg_types)
    for n,v in zip(fn.arg_names, input_values):
      self.value_env[n] = v
    self.visit_block(fn.body)

  def unify_scalar_var(self, x, y):
    """
    Unification is different than combining in that it imposes constraints on
    the program. If, for example, we're unifying some scalar that's reached the
    top of the lattice (and thus know nothing about it statically), with a
    scalar known to be some constant-- then the result is we expect both
    variables to be equal to that constant.
    """

    assert isinstance(x, Var), "Expected scalar variable, but got: " + str(x)
    assert isinstance(y, Scalar), "Expected scalar, but got: " + str(y)
    if y == any_scalar:
      return x
    equivs = self.equivalence_classes.get(x, set([]))
    equivs.add(y)
    for var in equivs:
      self.equivalence_classes[var] = equivs
    if isinstance(y, Const):
      for var in equivs:
        self.value_env[var] = y
      return y
    else:
      return var

  def unify_scalar_pairs(self, xs, ys):
    result_elts = []
    for xi, yi in zip(xs.elts, ys.elts):
      result_elts.append(self.unify_scalars(xi, yi))
    return result_elts

  def unify_scalar_list(self, values):
    assert len(values) > 0
    acc = any_scalar
    for v in values:
      acc = self.unify_scalars(acc, v)
    return acc

  def unify_scalars(self, x, y):
    if isinstance(x, Unknown):
      return y
    elif isinstance(y, Unknown):
      return x
    elif isinstance(x, Var):
      return self.unify_scalar_var(x, y)
    elif isinstance(y, Var):
      return self.unify_scalar_var(y, x)
    else:
      raise RuntimeError("Unsupported by unify: %s, %s" % (x,y))

  def visit_merge_loop_start(self, merge):
    for (k, (l, _)) in merge.iteritems():
      self.value_env[k] = self.visit_expr(l)
  
  def visit_merge_loop_repeat(self, merge):
    self.visit_merge(merge)

  def visit_merge(self, merge):
    for (k, (l,r)) in merge.iteritems():
      new_l = self.visit_expr(l)
      new_r = self.visit_expr(r)
      self.value_env[k] = new_l.combine(new_r)

  def visit_expr(self, expr):
    abstract_shape = SyntaxVisitor.visit_expr(self, expr)
    assert abstract_shape is not None, \
        "Unsupported expression in shape inference: %s" % expr.node_type()
    return abstract_shape

  def visit_Alloc(self, expr):
    return Ptr(any_scalar)

  def visit_Cast(self, expr):
    return any_scalar 
  
  def visit_TypeValue(self, expr):
    return unknown_value
  
  def visit_Struct(self, expr):
    if isinstance(expr.type, ArrayT):
      shape_tuple = self.visit_expr(expr.args[1])
      return make_shape(shape_tuple.elts)
    elif isinstance(expr.type, TupleT):
      return Tuple(self.visit_expr_list(expr.args))
    elif isinstance(expr.type, SliceT):
      start, stop, step = self.visit_expr_list(expr.args)
      return Slice(start, stop, step)
    else:
      return unknown_value

  def visit_Fn(self, fn):
    return Closure(fn, [])


  def visit_UntypedFn(self, fn):
    return Closure(fn, [])
  
  def visit_TypedFn(self, fn):
    return Closure(fn, [])

  def shape_from_tuple(self, expr):  
    shape_tuple = self.visit_expr(expr)
    if shape_tuple.__class__ is Tuple:
      return make_shape(tuple(shape_tuple.elts))
    elif shape_tuple.__class__ is Const:
      return make_shape((shape_tuple.value,))
    else:
      return make_shape( (any_scalar,) * expr.type.rank)

  def tuple_from_shape(self, expr):
    shape = self.visit_expr(expr)
    if shape.__class__ is Shape:
      return Tuple(tuple(shape.dims))
    elif shape.__class__ is Const:
      return Tuple( (shape.value,) )
    else:
      return Tuple( (any_scalar,) * expr.type.rank) 
   
  def visit_ArrayView(self, expr):
    return self.shape_from_tuple(expr.shape)
  
  def visit_Reshape(self, expr):
    return self.shape_from_tuple(expr.shape)
  
  def visit_Shape(self, expr):
    return self.tuple_from_shape(expr.array)
  
  def visit_Transpose(self, expr):
    shape = self.visit_expr(expr.array)
    if shape.__class__ is Shape:
      return Shape(tuple(reversed(shape.dims)))
    else:
      return shape 
  
  def visit_AllocArray(self, expr):
    return self.shape_from_tuple(expr.shape)
    
  def visit_Array(self, expr):
    elts = self.visit_expr_list(expr.elts)
    elt = combine_list(elts)
    n = len(elts)
    res = increase_rank(elt, 0, const(n))
    return res

  
  def visit_ConstArray(self, expr):
    return self.shape_from_tuple(expr.shape)
  
  def visit_ConstArrayLike(self, expr):
    return self.visit_expr(expr.array)
  
  def ravel(self, shape):        
    if isinstance(shape, Shape):
      nelts = const(1)
      for dim in shape.dims:
        nelts = self.mul(nelts, dim)
      return Shape((nelts,))
    else:
      return any_value 
  

  def visit_Ravel(self, expr):
    shape = self.visit_expr(expr.array)
    return self.ravel(shape)
  
  def visit_Range(self, expr):
    start = self.visit_expr(expr.start)
    stop = self.visit_expr(expr.stop)
    step = self.visit_expr(expr.step)
    slice_value = self.slice_value(start, stop, step)
    if slice_value.__class__ is ConstSlice:
      return Shape( (slice_value.nelts,))
    else:
      return Shape( (any_scalar,) )

  
  def visit_Slice(self, expr):
    step = self.visit_expr(expr.step)
    if expr.start.__class__ is syntax.Var and \
       expr.stop.__class__ is syntax.Var and \
       step.__class__ is Const:
      start_name = expr.start.name

      stop_name = expr.stop.name
      offsets = self.known_offsets.get(stop_name, [])
      step = step.value if step.value else 1
      for (other_var, offset) in offsets:

        if other_var == start_name:
          nelts = (offset + step - 1) /  step
          # assert False, (start_name, stop_name, offsets)
          return ConstSlice(nelts)

    start = self.visit_expr(expr.start)
    stop = self.visit_expr(expr.stop)
    return self.slice_value(start, stop, step)

  def visit_Const(self, expr):
    return Const(expr.value)

  def visit_ClosureElt(self, expr):
    clos = self.visit_expr(expr.closure)
    assert clos.__class__ is Closure, \
        "Unexpected closure shape %s for expression %s" % (clos, expr)
    return clos.args[expr.index]

  def visit_TupleProj(self, expr):
    t = self.visit_expr(expr.tuple)
    assert isinstance(t, Tuple), "Expected tuple type but got %s : %s" % (t, type(t))
    return t.elts[expr.index]

  def visit_Attribute(self, expr):
    v = self.visit_expr(expr.value)
    name = expr.name

    if v.__class__ is Shape:
      if name == 'shape':
        return Tuple(v.dims)
      elif name == 'strides':
        return Tuple((any_scalar,) * len(v.dims) )
      elif name in ('offset', 'size', 'nelts'):
        return any_scalar
      elif name == 'data':
        return Ptr(any_scalar)
    elif v.__class__ is Tuple:
      if name.startswith('elt'):
        idx = int(name[3:])
      else:
        idx = int(name)
      return v[idx]
    
    elif v.__class__ is Slice:
      return getattr(v, name)
    
    elif v.__class__ is Closure:
      if name.startswith('elt'):
        idx = int(name[3:])
      elif name.startswith('closure_elt'):
        idx = int(name[len('closure_elt'):])
      else:
        idx = int(name)
      return v.args[idx]
        
      
    elif v.__class__ is Struct:
      return v.values[v.fields.index(name)]

    t = expr.value.type.field_type(name)
    
    if isinstance(t, ScalarT):
      return any_scalar
    else:
      return any_value

  def visit_PrimCall(self, expr):
    
    p = expr.prim
    args = self.visit_expr_list(expr.args)
    if p == prims.add: 
      return self.add(args[0], args[1])
    elif p == prims.subtract:
      return self.sub(args[0], args[1])
    elif p == prims.multiply:
      return self.mul(args[0], args[1])
    elif p == prims.divide:
      return self.div(args[0], args[1])
    else:
      result = shape.combine_list(args, preserve_const = False)
      if result.__class__ is Shape:
        return result
      else:
        # once a scalar passes through some prim, it's not longer the same value!
        return any_scalar 

  def visit_Select(self, expr):
    cond = self.visit_expr(expr.cond)
    falseval = self.visit_expr(expr.true_value)
    trueval = self.visit_expr(expr.false_value)

    return cond.combine(falseval).combine(trueval)
    
  def visit_Var(self, expr):
    name = expr.name
    if name in self.value_env:
      return self.value_env[name]
    elif name in self.equivalence_classes:
      for other_name in self.equivalence_classes[name]:
        if other_name in self.value_env:
          return self.value_env[other_name]
    raise RuntimeError("Unknown variable: %s in function %s" %  (expr, self.fn.name))

  def visit_Tuple(self, expr):
    return Tuple(self.visit_expr_list(expr.elts))

  def visit_Call(self, expr):
    fn = self.visit_expr(expr.fn)
    args = self.visit_expr_list(expr.args)
    return symbolic_call(fn, args)

  def visit_Closure(self, clos):
    assert not isinstance(clos.fn, str), \
        "[ShapeInference] Function names in closures not supported: " + clos.fn
    fn = self.visit_expr(clos.fn)
    closure_arg_shapes = self.visit_expr_list(clos.args)
    if fn.__class__ is Closure:
      closure_arg_shapes = tuple(fn.args) + tuple(closure_arg_shapes)
      fn = fn.fn
    return Closure(fn, closure_arg_shapes)

  def visit_Index(self, expr):
    arr = self.visit_expr(expr.value)
    idx = self.visit_expr(expr.index)
    
    if arr.__class__ is Tuple and idx.__class__ is Const:
      return arr[idx.value]
    elif arr.__class__ is Shape:
      if isinstance(idx, Scalar):
        return shape.lower_rank(arr, 0)
      
      elif idx.__class__ is Shape:
        assert len(idx.dims) <= len(arr.dims), \
            "Can't index into rank %d array with rank %d indices" % \
            (len(arr.dims), len(idx.dims))
        dims = [d for d in arr.dims]
        for (i,d) in enumerate(idx.dims):
          dims[i] = d
        return shape.make_shape(dims)
      else:
        return self.index(arr, idx)
      
    elif arr.__class__ is Ptr:
      assert isinstance(arr.elt_shape, Scalar)
      assert isinstance(idx, Scalar)
      
      return any_scalar
    
    if isinstance(arr, Scalar):
      assert False, "Expected %s to be array, shape inference found scalar" % (arr,)
    elif arr == shape.any_value:
      raise ShapeInferenceFailure(expr, self.fn)
    assert False, \
        "Can't index (%s) with array shape %s and index shape %s" % \
        (expr, arr, idx)


  
  def visit_IndexMap(self, expr):
    bounds = self.visit_expr(expr.shape)
    clos = self.visit_expr(expr.fn)
    if isinstance(clos.fn.input_types[-1], TupleT) or bounds.__class__ is not Tuple:
      indices = [bounds]
    else:
      indices = bounds.elts 
    elt_result = symbolic_call(clos, indices)
    return make_shape(combine_dims(bounds, elt_result))
    
    
  def visit_IndexReduce(self, expr):
    fn = self.visit_expr(expr.fn)
    combine = self.visit_expr(expr.combine)
    bounds = self.visit_expr(expr.shape)
    if isinstance(fn.fn.input_types[-1], TupleT) or bounds.__class__ is not Tuple:
      indices = [bounds]
    else:
      indices = bounds.elts
    elt_shape = symbolic_call(fn, indices)
    init_shape = elt_shape if self.expr_is_none(expr.init) else self.visit_expr(expr.init) 
    return symbolic_call(combine, [init_shape, elt_shape])

  def visit_IndexScan(self, expr):
    fn = self.visit_expr(expr.fn)
    combine = self.visit_expr(expr.combine)
    emit = self.visit_expr(expr.emit)
    bounds = self.visit_expr(expr.shape)
    if isinstance(fn.fn.input_types[-1], TupleT) or bounds.__class__ is not Tuple:
      indices = [bounds]
    else:
      indices = bounds.elts
    elt_shape = symbolic_call(fn, indices)
    init_shape = elt_shape if self.expr_is_none(expr.init) else self.visit_expr(expr.init) 
    acc_shape = symbolic_call(combine, [init_shape, elt_shape])
    output_elt_shape = symbolic_call(emit, [acc_shape])
    return make_shape(combine_dims(bounds, output_elt_shape))


  def normalize_axes(self, axis, args):
    if isinstance(axis, Expr):
      axis = unwrap_constant(axis)
    if isinstance(axis,tuple):
      axes = axis 
    else:
      axes = (axis,) * len(args)
    
    assert len(axes) == len(args), \
      "Mismatch between args %s and axes %s" % (args, axis)
    return axes 
  
  def adverb_elt_shapes(self, arg_shapes, axes):
    """
    Slice into array shapes along the specified axis 
    """
    elt_shapes = []
    for arg_shape, axis in zip(arg_shapes, axes):
      if axis is None:
        elt_shapes.append(any_scalar)
      elif axis < self.rank(arg_shape):
        elt_shapes.append(self.slice_along_axis(arg_shape, axis))
      else:
        elt_shapes.append(arg_shape)
    return elt_shapes 

  def inner_map_result_shape(self, elt_result, arg_shapes, axes):
    max_rank = self.max_rank(arg_shapes)    
    for i, arg_shape in enumerate(arg_shapes):
      r = self.rank(arg_shape)
      if r == max_rank:
        axis = axes[i]
        if axis is None:
          combined_dims = dims(arg_shape) + dims(elt_result)
          if len(combined_dims) > 0:
            return Shape(combined_dims)
          else:
            return any_scalar 
        else:
          return increase_rank(elt_result, 0, arg_shape.dims[axis])
    return elt_result
    
  def outer_map_result_shape(self, elt_result, arg_shapes, axes):
    result_dims = list(dims(elt_result))
    for i, arg_shape in enumerate(arg_shapes):
      r = self.rank(arg_shape)
      if r > 0:
        axis = axes[i]
        if axis is None:
          result_dims.extend(arg_shape.dims)
        else:
          result_dims.append(arg_shape.dims[axis])
    return make_shape(result_dims)
    
  def visit_Map(self, expr):
    arg_shapes = self.visit_expr_list(expr.args)
    fn = self.visit_expr(expr.fn)
    axes = self.normalize_axes(expr.axis, expr.args)
    elt_shapes = self.adverb_elt_shapes(arg_shapes, axes)    
    elt_result = symbolic_call(fn, elt_shapes)
    return self.inner_map_result_shape(elt_result, arg_shapes, axes)
  
  def expr_is_none(self, expr):
    return expr is None or expr.type.__class__ is NoneT
  
  def visit_Reduce(self, expr):
    fn = self.visit_expr(expr.fn)
    combine = self.visit_expr(expr.combine)
    arg_shapes = self.visit_expr_list(expr.args)
    axes = self.normalize_axes(expr.axis, expr.args)
    elt_shapes = self.adverb_elt_shapes(arg_shapes, axes)
    elt_result = symbolic_call(fn, elt_shapes)
    init = elt_result if self.expr_is_none(expr.init) else self.visit_expr(expr.init) 
    return symbolic_call(combine, [init, elt_result])
      
  def visit_Scan(self, expr):
    fn = self.visit_expr(expr.fn)
    combine = self.visit_expr(expr.combine)
    arg_shapes = self.visit_expr_list(expr.args)
    axes = self.normalize_axes(expr.axis, expr.args)
    elt_shapes = self.adverb_elt_shapes(arg_shapes, axes)
    elt_result = symbolic_call(fn, elt_shapes)
    init = elt_result if self.expr_is_none(expr.init) else self.visit_expr(expr.init) 
    acc_shape = symbolic_call(combine, [init, elt_result])
    emit = self.visit_expr(expr.emit)
    emit_shape = symbolic_call(emit, [acc_shape])
    return self.inner_map_result_shape(emit_shape, arg_shapes, axes)
    
  def visit_OuterMap(self, expr):
    fn = self.visit_expr(expr.fn)
    arg_shapes = self.visit_expr_list(expr.args)
    axes = self.normalize_axes(expr.axis, expr.args)
    elt_shapes = self.adverb_elt_shapes(arg_shapes, axes)    
    elt_result = symbolic_call(fn, elt_shapes)
    return self.outer_map_result_shape(elt_result, arg_shapes, axes)

  def visit_Assign(self, stmt):
    rhs = self.visit_expr(stmt.rhs)
    if stmt.lhs.__class__ in (syntax.Var, syntax.Tuple):
      bind_syntax(stmt.lhs, rhs, self.value_env)
    
  def visit_Return(self, stmt):
    new_value = self.visit_expr(stmt.value)
    old_value = self.value_env.get("$return", unknown_value)
    combined = old_value.combine(new_value)
    self.value_env["$return"] = combined

  def visit_ForLoop(self, stmt):
    self.value_env[stmt.var.name] = any_scalar
    SyntaxVisitor.visit_ForLoop(self, stmt)
    # visit body a second time in case first-pass fixed values relative 
    # to initial value of iteration vars 
    self.visit_block(stmt.body)
  
  def visit_While(self, stmt):
    SyntaxVisitor.visit_While(self, stmt)
    # visit body a second time in case first-pass fixed values relative 
    # to initial value of iteration vars 
    self.visit_block(stmt.body)
    
    
_shape_env_cache = {}
def shape_env(typed_fn):
  key = typed_fn.cache_key
  
  if key in _shape_env_cache:
    return _shape_env_cache[key]

  shape_inference = ShapeInference()
  shape_inference.visit_fn(typed_fn)
  env = shape_inference.value_env
  _shape_env_cache[key] = env
  return env

_shape_cache = {}
def call_shape_expr(typed_fn):
  key = typed_fn.cache_key
  if key in _shape_cache:
    return _shape_cache[key]
  env = shape_env(typed_fn)
  abstract_shape = env.get("$return", Const(None))
  _shape_cache[key] = abstract_shape
  return abstract_shape

def bind_syntax(lhs, rhs, env):
  if isinstance(lhs, syntax.Tuple):
    assert isinstance(rhs, Tuple), "Expected tuple on RHS of binding %s = %s" % (lhs,rhs)
    for l,r in zip(lhs.elts, rhs.elts):
      bind_syntax(l, r, env)
  elif isinstance(lhs, syntax.Var):
    env[lhs.name] = rhs

def bind(lhs, rhs, env):
  if isinstance(lhs, Var):
    env[lhs] = rhs
  elif isinstance(lhs, Shape):
    assert isinstance(rhs, Shape), "Expected %s, got %s" % (lhs, rhs)
    bind_pairs(lhs.dims, rhs.dims, env)
  elif isinstance(lhs, Closure):
    assert isinstance(rhs, Closure)
    bind_pairs(lhs.args, rhs.args, env)
  elif isinstance(lhs, Tuple):
    if rhs == any_value: 
      bind_pairs(lhs.elts, [any_value for _ in lhs.elts], env)
    elif lhs == unknown_value:
      bind_pairs(lhs.elts, [unknown_value for _ in lhs.elts], env)
    else:
      assert isinstance(rhs, Tuple), "Expected tuple on RHS of binding %s = %s" % (lhs,rhs)
      bind_pairs(lhs.elts, rhs.elts, env)
  else:
    raise RuntimeError("Unexpected shape LHS: %s" % lhs)

def bind_pairs(xs, ys, env):
  assert len(xs) == len(ys), \
      "Can't bind %s and %s due to unequal lengths" % (xs, ys)
  for (x,y) in zip(xs,ys):
    bind(x,y,env)

def subst(x, env):
  if isinstance(x, Var):
    assert x in env, "Unknown variable %s" % x
    return env[x]
  elif isinstance(x, Scalar):
    return x
  elif isinstance(x, Shape):
    return make_shape(subst_list(x.dims, env))
  elif isinstance(x, Tuple):
    return Tuple(tuple((subst_list(x.elts, env))))
  elif isinstance(x, Closure):
    return Closure(x.fn, subst_list(x.args, env))
  elif isinstance(x, Ptr):
    return Ptr(subst(x.elt_shape, env))
  else:
    raise RuntimeError("Unexpected abstract expression: %s" % x)

def subst_list(xs, env):
  return [subst(x, env) for x in xs]

def symbolic_call(fn, abstract_inputs):
  # result in terms of variables like input0, (shape: input1, input2), etc..
  if fn.__class__ is Closure:
    closure_elts = tuple(fn.args)
    fn = fn.fn
  else:
    closure_elts = ()

  abstract_result_value = call_shape_expr(fn)
  conv = shape_from_type.Converter()
  shape_formals = conv.from_types(fn.input_types)
  
  env = {}
  bind_pairs(shape_formals, closure_elts + tuple(abstract_inputs), env)

  return subst(abstract_result_value, env)



from parakeet.testing_helpers import expect, run_local_tests

def call_const_lambda():
  return (lambda: 1)()

def test_call_const_lambda():
  expect(call_const_lambda, [], 1)

def call_identity_lambda(x):
  return (lambda y: y)(x)

def test_call_identity_lambda():
  expect(call_identity_lambda, [1], 1)

def call_closure_lambda(x):
  return (lambda y: x + y)(x)

def test_call_closure_lambda():
  expect(call_closure_lambda, [1], 2)

if __name__ == "__main__":
  run_local_tests()





from parakeet import Int64, make_tuple_type, make_array_type
from parakeet.analysis.escape_analysis import EscapeAnalysis
from parakeet.syntax import Tuple, Var, TypedFn, Assign, Return, Array, TupleProj, zero_i64, one_i64
from parakeet.testing_helpers import run_local_tests   

 

array_t = make_array_type(Int64, 1)
tuple_t = make_tuple_type((Int64, array_t))
nested_tuple_t = make_tuple_type((tuple_t, tuple_t))
array_const = Array([one_i64], type = array_t)

a_int = Var("a_int", type = Int64)
b_array = Var("b_array", type = array_t)
c_tuple = Var("c_tuple", type = tuple_t)
d_tuple = Var("d_tuple", type = tuple_t)
e_nested_tuple = Var("e_nested_tuple", type= nested_tuple_t)
f_nested_tuple = Var("f_nested_tuple", type = nested_tuple_t)
body = [
  Assign(a_int, one_i64),  
  Assign(b_array, array_const), 
  Assign(c_tuple, Tuple(elts = (a_int, b_array), type = tuple_t)),
  Assign(d_tuple, Tuple(elts = (a_int, array_const), type = tuple_t)),
  Assign(e_nested_tuple, Tuple(elts = (c_tuple, c_tuple), type = nested_tuple_t)),
  Assign(f_nested_tuple, Tuple(elts = (d_tuple, d_tuple), type = nested_tuple_t)),  
  Return(b_array) 
]
tenv = { "a_int": Int64, 
         "b_array": array_t, 
         "c_tuple": tuple_t, 
         "d_tuple": nested_tuple_t, 
         "e_nested_tuple": nested_tuple_t,
         "f_nested_tuple" : nested_tuple_t, 
        }

fn = TypedFn(name = "test_escape_analysis", 
             type_env = tenv, 
             input_types = (), 
             arg_names = (), 
             body = body, 
             return_type = nested_tuple_t
            )

def test_escape_analysis():
  escape_analysis = EscapeAnalysis()
  escape_analysis.visit_fn(fn)
  may_escape =  escape_analysis.may_escape
 
  assert "a_int" not in may_escape, "Scalars don't count as escaping"
  assert "b_array" in may_escape
  assert "c_tuple" in may_escape
  assert "d_tuple" not in may_escape
  assert "e_nested_tuple" in may_escape 
  assert "f_nested_tuple" not in may_escape
  
if __name__ == '__main__':
  run_local_tests()




import llvm.core as llc 
from llvm.core import Builder, ATTR_NO_CAPTURE #, Module 
from llvm.core import Type as lltype

from .. import config, prims, syntax 
from .. analysis import may_escape
from .. ndtypes import BoolT, FloatT, SignedT, UnsignedT, ScalarT, NoneT, Float64
from .. ndtypes import Int32, Int64, PtrT, Bool 
from .. syntax import Var, Struct, Index, TypedFn, Attribute 

import llvm_config 
import llvm_context
import llvm_convert
import llvm_types
import llvm_prims
from llvm_helpers import const, int32, zero 
from llvm_types import llvm_value_type, llvm_ref_type
from parakeet.llvm_backend.llvm_convert import to_bit, from_bit


_escape_analysis_cache = {}
class Compiler(object):
  def __init__(self, fundef, llvm_cxt = llvm_context.global_context):
    self.parakeet_fundef = fundef
    if config.opt_stack_allocation:
      self.may_escape = may_escape(fundef)
    else:
      self.may_escape = None
    self.llvm_context = llvm_cxt
    self.vars = {}
    self.initialized = set([])
    # Initializes the variables dictionary and returns a builder object
    llvm_input_types = map(llvm_ref_type, fundef.input_types)
    llvm_output_type = llvm_ref_type(fundef.return_type)
    llvm_fn_t = lltype.function(llvm_output_type, llvm_input_types)

    self.llvm_fn = self.llvm_context.module.add_function(llvm_fn_t, fundef.name)

    for arg in self.llvm_fn.args:
      if not llvm_types.is_scalar(arg.type):
        arg.add_attribute(ATTR_NO_CAPTURE)

    self.llvm_fn.does_not_throw = True

    self.entry_block, self.entry_builder = self.new_block("entry")
    self._init_vars(self.parakeet_fundef, self.entry_builder)

  def new_block(self, name):
    bb = self.llvm_fn.append_basic_block(name)
    builder = Builder.new(bb)
    return bb, builder

  def _init_vars(self, fundef, builder):
    """
    Create a mapping from variable names to stack locations, these will later be
    converted to SSA variables by the mem2reg pass.
    """
    n_expected = len(fundef.arg_names)
    n_compiled = len(self.llvm_fn.args)
    assert n_compiled == n_expected, \
        "Expected %d args (%s) but compiled code had %d args (%s)" % \
        (n_expected, fundef.arg_names, n_compiled, self.llvm_fn.args)

    for (name, t) in fundef.type_env.iteritems():
      if not name.startswith("$"):
        llvm_t = llvm_ref_type(t)
        stack_val = builder.alloca(llvm_t, name)
        self.vars[name] = stack_val

    for llvm_arg, name in zip(self.llvm_fn.args, fundef.arg_names):
      self.initialized.add(name)
      llvm_arg.name = name
      if name in self.vars:
        builder.store(llvm_arg, self.vars[name])

  def attribute_lookup(self, struct, name, builder):
    """
    Helper for getting the address of an attribute lookup, used both when
    setting and getting attributes
    """
    llvm_struct = self.compile_expr(struct, builder)
    struct_t = struct.type
    field_pos = struct_t.field_pos(name)
    field_type = struct_t.field_type(name)
    indices = [int32(0), int32(field_pos)]
    ptr_name = "%s_ptr" % name
    ptr = builder.gep(llvm_struct, indices, ptr_name)
    return ptr, field_type

  def compile_Var(self, expr, builder):
    name = expr.name
    assert name in self.initialized, "%s uninitialized" % name
    ref = self.vars[expr.name]
    val = builder.load(ref, expr.name + "_val")
    return val

  def compile_Const(self, expr, builder):
    t = expr.type

    if isinstance(t, NoneT):
      return const(0, Int64)
    else:
      assert isinstance(expr.type, ScalarT), \
          "Expected scalar constant but got %s" % expr.type
    return const(expr.value, expr.type)

  def compile_Cast(self, expr, builder):
    llvm_value = self.compile_expr(expr.value, builder)
    return llvm_convert.convert(llvm_value, expr.value.type, expr.type, builder)

  def compile_Struct(self, expr, builder, local = False):
    struct_t = expr.type
    llvm_struct_t = llvm_value_type(struct_t)
    name = expr.type.node_type()
    if local:
      struct_ptr = builder.alloca(llvm_struct_t, name + "_local_ptr")
    else:
      struct_ptr = builder.malloc(llvm_struct_t, name + "_ptr")

    for (i, elt) in enumerate(expr.args):
      field_name, field_type = struct_t._fields_[i]
      assert elt.type == field_type, \
          "Mismatch between expected type %s and given %s for field '%s' " % \
          (field_type, elt.type, field_name)
      elt_ptr = builder.gep(struct_ptr, [int32(0), int32(i)], "field%d_ptr" % i)
      llvm_elt = self.compile_expr(elt, builder)
      builder.store(llvm_elt, elt_ptr)

    return struct_ptr

  def compile_Alloc(self, expr, builder):
    elt_t = expr.elt_type
    llvm_elt_t = llvm_types.llvm_value_type(elt_t)
    n_elts = self.compile_expr(expr.count, builder)
    return builder.malloc_array(llvm_elt_t, n_elts, "data_ptr")

  def compile_Index(self, expr, builder):
    llvm_arr = self.compile_expr(expr.value, builder)
    llvm_index = self.compile_expr(expr.index, builder)
    pointer = builder.gep(llvm_arr, [llvm_index], "elt_pointer")
    elt = builder.load(pointer, "elt", align = 16) #, invariant = True)
    return elt
  
  def compile_Select(self, expr, builder):
    cond = self.compile_expr(expr.cond, builder)
    cond = llvm_convert.to_bit(cond, builder)
    trueval = self.compile_expr(expr.true_value, builder)
    falseval = self.compile_expr(expr.false_value, builder)
    result =  builder.select(cond, trueval, falseval, "select_result")
    return result 
  
  def compile_Attribute(self, expr, builder):
    field_ptr, _ = \
        self.attribute_lookup(expr.value, expr.name, builder)
    field_value = builder.load(field_ptr, "%s_value" % expr.name)
    return field_value

  def compile_TypedFn(self, expr, builder):
    (target_fn, _, _) = compile_fn(expr)
    return target_fn

  def compile_Call(self, expr, builder):
    assert expr.fn.__class__ is TypedFn
    typed_fundef = expr.fn
    (target_fn, _, _) = compile_fn(typed_fundef)
    arg_types = syntax.get_types(expr.args)
    llvm_args = [self.compile_expr(arg, builder) for arg in expr.args]
    assert len(arg_types) == len(llvm_args)
    return builder.call(target_fn, llvm_args, 'call_result')

  def cmp(self, prim, t, llvm_x, llvm_y, builder, result_name = None):
    if result_name is None:
      result_name = prim.name + "_result"

    if isinstance(t, FloatT):
      cmp_op = llvm_prims.float_comparisons[prim]
      return builder.fcmp(cmp_op, llvm_x, llvm_y, result_name)
    elif isinstance(t, SignedT):
      cmp_op = llvm_prims.signed_int_comparisons[prim]
      return builder.icmp(cmp_op, llvm_x, llvm_y, result_name)
    else:
      assert isinstance(t, (BoolT, UnsignedT)), "Unexpected type for comparison %s: %s" % (prim, t)
      cmp_op = llvm_prims.unsigned_int_comparisons[prim]
      return builder.icmp(cmp_op, llvm_x, llvm_y, result_name)

  
  def lt(self, t, llvm_x, llvm_y, builder, result_name = None ):
    return self.cmp(prims.less, t, llvm_x, llvm_y, builder, result_name)
  
  def lte(self, t, llvm_x, llvm_y, builder, result_name = None ):
    return self.cmp(prims.less_equal, t, llvm_x, llvm_y, builder, result_name)
  
  def neq(self, t, llvm_x, llvm_y, builder, result_name = None):
    return self.cmp(prims.not_equal, t, llvm_x, llvm_y, builder, result_name)
  
  def sub(self, t, x, y, builder, result_name = "sub"):
    if isinstance(t, FloatT):
      return builder.fsub(x, y, result_name)
    else:
      return builder.sub(x, y, result_name)
    
  
  def add(self, t, x, y, builder, result_name = "add"):
    if isinstance(t, FloatT):
      return builder.fadd(x, y, result_name)
    else:
      return builder.add(x, y, result_name)
  
  
  def mul(self, t, x, y, builder, result_name = "mul"):
    if isinstance(t, FloatT):
      return builder.fmul(x, y, result_name)
    else:
      return builder.mul(x, y, result_name)
  
    
  def neg(self, x, builder):
    if isinstance(x.type, llc.IntegerType):
      return builder.neg(x, "neg")
    else:
      return builder.fsub(zero(x.type), x, "neg") 
    
  def prim(self, prim, t, llvm_args, builder, result_name = None):
    if result_name is None:
      result_name = prim.name + "_result"

    if isinstance(prim, prims.Cmp):
      bit = self.cmp(prim, t, llvm_args[0], llvm_args[1], builder)
      return llvm_convert.to_bool(bit,builder)
    
    elif prim == prims.maximum:
      x, y = llvm_args
      bit = self.cmp(prims.greater_equal, t, x, y, builder)
      return builder.select(bit, x, y)
    
    elif prim == prims.minimum:
      x,y = llvm_args
      bit = self.cmp(prims.less_equal, t, x, y, builder)
      return builder.select(bit, x, y)
    
    elif prim == prims.negative:
      if t == Bool: 
        bit = llvm_convert.to_bit(llvm_args[0], builder)
        negated = builder.not_(bit)
        return llvm_convert.to_bool(negated, builder)
      return self.neg(llvm_args[0], builder)
    
    # python's remainder is weird in that it preserve's the sign of 
    # the second argument, whereas LLVM's srem/frem operators preserve
    # the sign of the first 
    elif prim == prims.mod:
      x,y = llvm_args 
      if isinstance(t, (UnsignedT, BoolT)):
        return builder.urem(llvm_args[0], llvm_args[1], "modulo")
      elif isinstance(t, SignedT): 
        rem = builder.srem(x,y, "modulo")
      else:
        assert isinstance(t, FloatT)
        rem = builder.frem(llvm_args[0], llvm_args[1], "modulo")

      y_is_negative = self.lt(t, y, zero(y.type), builder, "second_arg_negative")
      rem_is_negative = self.lt(t, rem, zero(rem.type), builder, "rem_is_negative")
      y_nonzero = self.neq(t, y, zero(y.type), builder, "second_arg_nonzero")
      rem_nonzero = self.neq(t, rem, zero(x.type), builder, "rem_nonzero")
      neither_zero = builder.and_(y_nonzero, rem_nonzero, "neither_zero")
      diff_signs = builder.xor(y_is_negative, rem_is_negative, "different_signs")
      should_flip = builder.and_(neither_zero, diff_signs, "should_flip") 
      flipped_rem = self.add(t, y, rem, builder, "flipped_rem")
      return builder.select(should_flip, flipped_rem, rem)

    elif prim == prims.power:
      x,y = llvm_args
       
      if isinstance(t, FloatT):
        new_t = t 
      else:
        new_t = Float64
      x = llvm_convert.convert(x, t, new_t, builder)
      y = llvm_convert.convert(y, t, new_t, builder)
      llvm_op = llvm_prims.get_float_op(prim, new_t)
      result = builder.call(llvm_op, [x,y])
      return llvm_convert.convert(result, new_t, t, builder)
        
    elif isinstance(prim, prims.Arith) or isinstance(prim, prims.Bitwise):
      if isinstance(t, FloatT):
        instr = llvm_prims.float_binops[prim]
      elif isinstance(t, SignedT):
        instr = llvm_prims.signed_binops[prim]
      elif isinstance(t, UnsignedT):
        instr = llvm_prims.unsigned_binops[prim]
      else:
        assert isinstance(t, BoolT)
        instr = llvm_prims.bool_binops[prim]
      op = getattr(builder, instr)
      return op(name = result_name, *llvm_args)

    elif isinstance(prim, prims.Logical):
      if prim == prims.logical_and:
        result = builder.and_(name = result_name, 
                            lhs = to_bit(llvm_args[0], builder), 
                            rhs = to_bit(llvm_args[1], builder))
        return from_bit(result, t, builder)
      elif prim == prims.logical_not:
        result = builder.not_(to_bit(llvm_args[0], builder), name = result_name)
        return from_bit(result, t, builder)
      else:
        assert prim == prims.logical_or
        result = builder.or_(lhs = to_bit(llvm_args[0], builder), 
                           rhs = to_bit(llvm_args[1], builder), name = result_name)
        return from_bit(result, t, builder)
      
    elif prim == prims.abs:
      x = llvm_args[0]
      bit = self.cmp(prims.greater_equal, t,  x, zero(x.type), builder, "gt_zero")
      neg_value = self.neg(x, builder)
      return builder.select(bit, x, neg_value)
    elif isinstance(prim, prims.Float): 
      llvm_op = llvm_prims.get_float_op(prim, t)
      return builder.call(llvm_op, llvm_args)
    
    else:
      assert False, "UNSUPPORTED PRIMITIVE: %s" % prim

  def compile_PrimCall(self, expr, builder):
    args = expr.args
    # type specialization should have made types of arguments uniform,
    # so we only need to check the type of the first arg
    t = args[0].type
    llvm_args = [self.compile_expr(arg, builder) for arg in args]
    return self.prim(expr.prim, t, llvm_args, builder)

  def compile_expr(self, expr, builder):
    method_name = "compile_" + expr.node_type()
    return getattr(self, method_name)(expr, builder)

  def compile_Assign(self, stmt, builder):
    rhs_t = stmt.rhs.type
    # special case for locally allocated structs
    if self.may_escape is not None and \
       stmt.lhs.__class__ is Var and \
       stmt.rhs.__class__ is Struct and \
       stmt.lhs.name  not in self.may_escape:
      value = self.compile_Struct(stmt.rhs, builder, local = True)
    else:
      value = self.compile_expr(stmt.rhs, builder)
    if stmt.lhs.__class__ is Var:
      name = stmt.lhs.name
      lhs_t = stmt.lhs.type
      self.initialized.add(name)
      ref = self.vars[name]
    elif stmt.lhs.__class__ is Index:
      ptr_t = stmt.lhs.value.type
      assert isinstance(ptr_t, PtrT), \
          "Expected pointer, got %s" % ptr_t
      lhs_t = ptr_t.elt_type
      base_ptr = self.compile_expr(stmt.lhs.value, builder)
      index = self.compile_expr(stmt.lhs.index, builder)
      index = llvm_convert.from_signed(index, Int32, builder)
      ref = builder.gep(base_ptr, [index], "elt_ptr")
    else:
      assert stmt.lhs.__class__ is Attribute, \
          "Unexpected LHS: %s" % stmt.lhs
      struct = stmt.lhs.value
      ref, lhs_t = self.attribute_lookup(struct, stmt.lhs.name, builder)

    assert lhs_t == rhs_t, \
        "Type mismatch between LHS %s and RHS %s" % (lhs_t, rhs_t)

    builder.store(value, ref)
    return builder, False

  def compile_ExprStmt(self, stmt, builder):
    self.compile_expr(stmt.value, builder)
    return builder, False

  def compile_Return(self, stmt, builder):
    ret_val = self.compile_expr(stmt.value, builder)
    builder.ret(ret_val)
    return builder, True

  def compile_merge_left(self, phi_nodes, builder):
    for name, (left, _) in phi_nodes.iteritems():
      ref = self.vars[name]
      self.initialized.add(name)
      value = self.compile_expr(left, builder)
      builder.store(value, ref)

  def compile_merge_right(self, phi_nodes, builder):
    for name, (_, right) in phi_nodes.iteritems():
      ref = self.vars[name]
      self.initialized.add(name)
      value = self.compile_expr(right, builder)
      builder.store(value, ref)

  def compile_ForLoop(self, stmt, builder):
    # first compile the starting, boundary, and
    # increment values for the loop counter
    start = self.compile_expr(stmt.start, builder)
    stop = self.compile_expr(stmt.stop, builder)
    step = self.compile_expr(stmt.step, builder)

    # get the memory slot associated with the loop counter
    loop_var = self.vars[stmt.var.name]

    builder.store(start, loop_var)
    self.initialized.add(stmt.var.name)

    # ...and we'll need its Parakeet type later on
    # for calls to 'cmp' and 'prim'
    loop_var_t = stmt.var.type

    # any phi-bound variables should be initialized to their
    # starting values
    self.compile_merge_left(stmt.merge, builder)

    loop_bb, body_start_builder = self.new_block("loop_body")
    after_bb, after_builder = self.new_block("after_loop")

    # WARNING: Assuming loop is always increasing,
    # only enter the loop if we're less than the stopping value
    enter_cond = self.cmp(prims.less, loop_var_t,  start, stop,
                          builder, "enter_cond")
    builder.cbranch(enter_cond, loop_bb, after_bb)

    # TODO: what should we do if the body always ends in a return statement?
    body_end_builder, body_always_returns = \
      self.compile_block(stmt.body, body_start_builder)

    counter_at_end = body_end_builder.load(loop_var)
    # increment the loop counter
    incr = self.prim(prims.add, loop_var_t,
                     [counter_at_end, step],
                     body_end_builder,  "incr_loop_var")
    body_end_builder.store(incr, loop_var)
    self.compile_merge_right(stmt.merge, body_end_builder)

    exit_cond = self.cmp(prims.less, loop_var_t, incr, stop,
                         body_end_builder, "exit_cond")
    body_end_builder.cbranch(exit_cond, loop_bb, after_bb)
    # WARNING: what if the loop doesn't run? Should
    # we still be returning 'body_always_returns'?
    return after_builder, body_always_returns

  def compile_While(self, stmt, builder):
    # current flow ----> loop --------> exit--> after
    #    |                       skip------------|
    #    |----------------------/

    self.compile_merge_left(stmt.merge, builder)
    loop_bb, body_start_builder = self.new_block("loop_body")

    after_bb, after_builder = self.new_block("after_loop")
    enter_cond = self.compile_expr(stmt.cond, builder)
    enter_cond = llvm_convert.to_bit(enter_cond, builder)
    builder.cbranch(enter_cond, loop_bb, after_bb)

    body_end_builder, body_always_returns = \
        self.compile_block(stmt.body, body_start_builder)
    if not body_always_returns:
      exit_bb, exit_builder = self.new_block("loop_exit")
      self.compile_merge_right(stmt.merge, body_end_builder)
      repeat_cond = self.compile_expr(stmt.cond, body_end_builder)
      repeat_cond = llvm_convert.to_bit(repeat_cond, body_end_builder)
      body_end_builder.cbranch(repeat_cond, loop_bb, exit_bb)
      exit_builder.branch(after_bb)

    return after_builder, False

  def compile_If(self, stmt, builder):
    cond = self.compile_expr(stmt.cond, builder)
    cond = llvm_convert.to_bit(cond, builder)

    if len(stmt.true) == 0 and len(stmt.false) == 0:
      # if nothing happens in the loop bodies, just
      # emit select instructions
      for (name, (true_expr, false_expr)) in stmt.merge.iteritems():
        ref = self.vars[name]
        self.initialized.add(name)
        true_val = self.compile_expr(true_expr, builder)
        false_val = self.compile_expr(false_expr, builder)
        select_val = builder.select(cond, true_val, false_val)
        builder.store(select_val, ref)
      return builder, False
    else:
      # compile the two possible branches as distinct basic blocks
      # and then wire together the control flow with branches
      true_bb, true_builder = self.new_block("if_true")
      after_true, true_always_returns = \
        self.compile_block(stmt.true, true_builder)

      false_bb, false_builder = self.new_block("if_false")
      after_false, false_always_returns = \
          self.compile_block(stmt.false, false_builder)

      builder.cbranch(cond, true_bb, false_bb)

      # compile phi nodes as assignments and then branch
      # to the continuation block
      self.compile_merge_left(stmt.merge, after_true)
      self.compile_merge_right(stmt.merge, after_false)

      # if both branches return then there is no point
      # making a new block for more code
      # did both branches end in a return?
      both_always_return = true_always_returns and false_always_returns
      if both_always_return:
        return None, True
      after_bb, after_builder = self.new_block("if_after")
      if not true_always_returns:
        after_true.branch(after_bb)
      if not false_always_returns:
        after_false.branch(after_bb)
      return after_builder, False

  def compile_Comment(self, stmt, builder):
    return builder, False

  def compile_stmt(self, stmt, builder):
    """
    Translate an SSA statement into LLVM. Every translation function returns a
    builder pointing to the end of the current basic block and a boolean
    indicating whether every branch of control flow in that statement ends in a
    return. The latter is needed to avoid creating empty basic blocks, which
    were causing some mysterious crashes inside LLVM.
    """
    method_name = "compile_" + stmt.node_type()
    return getattr(self, method_name)(stmt, builder)

  def compile_block(self, stmts, builder):
    for stmt in stmts:
      builder, always_returns = self.compile_stmt(stmt, builder)
      if always_returns:
        return builder, always_returns
    return builder, False

  def compile_body(self, body):
    return self.compile_block(body, builder = self.entry_builder)
  
  
import os
from collections import namedtuple   
CompiledFn = namedtuple('CompiledFn', ('llvm_fn', 'llvm_exec_engine', 'parakeet_fn')) 

compiled_functions = {}
def compile_fn(fundef):
  key = fundef.cache_key 
  if key in compiled_functions:
    return compiled_functions[key]
  
  if config.print_lowered_function:
    print
    print "=== Lowered function ==="
    print
    print repr(fundef)
    print

  compiler = Compiler(fundef)
  compiler.compile_body(fundef.body)
  if llvm_config.print_unoptimized_llvm:
    print "=== LLVM before optimizations =="
    print
    print compiler.llvm_context.module
    print
  compiler.llvm_context.run_passes(compiler.llvm_fn)

  if config.print_generated_code:
    print "=== LLVM after optimizations =="
    print
    print compiler.llvm_context.module
    print

  if llvm_config.print_x86:
    print "=== Generated assembly =="
    print
    start_printing = False
    w,r = os.popen2("llc")
    w.write(str(compiler.llvm_context.module))
    w.close()
    assembly_str = r.read()
    r.close()
    for l in assembly_str.split('\n'):
      if l.strip().startswith('.globl'):
        if start_printing:
          break
        elif fundef.name in l:
          start_printing = True
      if start_printing:
        print l

  result = CompiledFn(compiler.llvm_fn,
                      compiler.llvm_context.exec_engine, 
                      fundef)
  compiled_functions[key] = result
  return result



import numpy as np
import time 

import parakeet 
from parakeet import testing_helpers

def sqr_dist(x,y):
  return sum( (x-y) ** 2)

def allpairs_dist_adverb(X,Y):
  return parakeet.allpairs(sqr_dist, X, Y)

def allpairs_dist_comprehensions_external(X,Y):
  return np.array([[sqr_dist(x,y) for y in Y] for x in X])

def allpairs_dist_comprehensions_internal(X,Y):
  def local_sqr_dist(x,y):
    return np.sum( (x-y)**2 )
  return np.array([[local_sqr_dist(x,y) for y in Y] for x in X])

m = 2
n = 3
d = 5 
X = np.random.randn(m,d)
Y = np.random.randn(n,d)
python_dists = np.array([[sqr_dist(x,y) for y in Y] for x in X])

def test_dists_adverb():
  testing_helpers.expect(allpairs_dist_adverb, [X,Y], python_dists)

def test_dists_comprehensions_external():
  testing_helpers.expect(allpairs_dist_comprehensions_external, [X,Y], python_dists)
  
def test_dists_comprehensions_internal():
  testing_helpers.expect(allpairs_dist_comprehensions_external, [X,Y], python_dists)

if __name__ == '__main__':
  testing_helpers.run_local_tests()



from numpy import sin, cos, dot, array 


def rotate(phi, theta, orig): 
  """
  This function rotates the point at orig about axis u(depends on phi) by the angle theta.  
  orig and output are in Cartesian coordinates
  """
  
  u = (-sin(phi), cos(phi), 0.0)
  rotM = array([
                 [cos(theta)+u[0]**2*(1-cos(theta)), u[0]*u[1]*(1-cos(theta)), u[1]*sin(theta)],
                 [u[0]*u[1]*(1-cos(theta)), cos(theta)+u[1]**2*(1-cos(theta)), -u[0]*sin(theta)],
                 [-u[1]*sin(theta), u[0]*sin(theta), cos(theta)]
                ])
  rotP = dot(rotM,orig)
  return rotP


from parakeet import testing_helpers
import numpy as np 

def test_rotate():
  x = np.arange(3) / 2.0
  for phi in (0, 0.25):
    for theta in (0, 0.25):
      testing_helpers.expect(rotate, [phi, theta, x], rotate(phi, theta, x))

if __name__ == "__main__":
  testing_helpers.run_local_tests()



from .. syntax import (Expr, Assign, ExprStmt, ForLoop, If, Return, While, Comment, ParFor, 
                       TypedFn, UntypedFn,  Closure, ClosureElt, Select,  
                       Attribute, Const, Index, PrimCall, Tuple, Var, 
                       Alloc, Array, Call, Struct, Shape, Strides, Range, Ravel, Transpose,
                       AllocArray, ArrayView, Cast, Slice, TupleProj, TypeValue,  
                       Map, Reduce, Scan, OuterMap, IndexMap, IndexReduce, IndexScan )

class SyntaxVisitor(object):
  """
  Traverse the statement structure of a syntax block, optionally collecting
  values
  """

  def visit_if_expr(self, expr):
    if isinstance(expr, Expr):
      self.visit_expr(expr)
      
  def visit_Var(self, expr):
    pass

  def visit_Const(self, expr):
    pass
  
  def visit_Fn(self, expr):
    pass 
  
  def visit_ClosureElt(self, expr):
    self.visit_expr(expr.closure)
  
  def visit_Closure(self, expr):
    self.visit_expr(expr.fn)
    self.visit_expr_list(expr.args)
    
    
  def visit_Tuple(self, expr):
    for elt in expr.elts:
      self.visit_expr(elt)

  def visit_PrimCall(self, expr):
    for arg in expr.args:
      self.visit_expr(arg)

  def visit_Attribute(self, expr):
    self.visit_expr(expr.value)

  def visit_Index(self, expr):
    self.visit_expr(expr.value)
    self.visit_expr(expr.index)

  def visit_UntypedFn(self, expr):
    pass 
  
  def visit_TypedFn(self, expr):
    pass

  def visit_Alloc(self, expr):
    self.visit_expr(expr.count)

  def visit_Struct(self, expr):
    for arg in expr.args:
      self.visit_expr(arg)

  def visit_Array(self, expr):
    for elt in expr.elts:
      self.visit_expr(elt)
      
  def visit_ArrayView(self, expr):
    self.visit_expr(expr.data)
    self.visit_expr(expr.shape)
    self.visit_expr(expr.strides)
    self.visit_expr(expr.offset)
    self.visit_expr(expr.size)

  def visit_AllocArray(self, expr):
    self.visit_expr(expr.shape)

  def visit_Ravel(self, expr):
    self.visit_expr(expr.array)

  def visit_Reshape(self, expr):
    self.visit_expr(expr.array)
    self.visit_expr(expr.shape)

  def visit_Shape(self, expr):
    self.visit_expr(expr.array)
  
  def visit_ConstArray(self, expr):
    self.visit_expr(expr.shape)
    self.visit_expr(expr.value)
    
  def visit_ConstArrayLike(self, expr):
    self.visit_expr(expr.array)
    self.visit_expr(expr.value)
  
  def visit_Slice(self, expr):
    self.visit_expr(expr.start)
    self.visit_expr(expr.stop)
    self.visit_expr(expr.step)

  def visit_Transpose(self, expr):
    self.visit_expr(expr.array)
    
  def visit_IndexMap(self, expr):
    self.visit_expr(expr.fn)
    self.visit_expr(expr.shape)

  def visit_IndexReduce(self, expr):
    self.visit_expr(expr.fn)
    self.visit_expr(expr.combine)
    self.visit_expr(expr.shape)
    self.visit_expr(expr.init)

  def visit_IndexScan(self, expr):
    self.visit_expr(expr.fn)
    self.visit_expr(expr.combine)
    self.visit_expr(expr.shape)
    self.visit_expr(expr.init)

  def visit_Map(self, expr):
    self.visit_expr(expr.fn)
    self.visit_if_expr(expr.axis)
    for arg in expr.args:
      self.visit_expr(arg)

  def visit_OuterMap(self, expr):
    self.visit_expr(expr.fn)
    self.visit_if_expr(expr.axis)
    for arg in expr.args:
      self.visit_expr(arg)

  def visit_Reduce(self, expr):
    self.visit_expr(expr.fn)
    self.visit_expr(expr.combine)
    self.visit_if_expr(expr.axis)
    self.visit_if_expr(expr.init)
    for arg in expr.args:
      self.visit_expr(arg)

  def visit_Scan(self, expr):
    self.visit_expr(expr.fn)
    self.visit_expr(expr.combine)
    self.visit_if_expr(expr.axis)
    self.visit_if_expr(expr.init)
    for arg in expr.args:
      self.visit_expr(arg)
      
  def visit_TupleProj(self, expr):
    return self.visit_expr(expr.tuple)

  def visit_Call(self, expr):
    self.visit_expr(expr.fn)
    for arg in expr.args:
      self.visit_expr(arg)

  def visit_Cast(self, expr):
    return self.visit_expr(expr.value)

  def visit_Range(self, expr):
    self.visit_expr(expr.start)
    self.visit_expr(expr.stop)
    self.visit_expr(expr.step)
  
  def visit_Select(self, expr):
    self.visit_expr(expr.cond)
    self.visit_expr(expr.true_value)
    self.visit_expr(expr.false_value)
    
  def visit_TypeValue(self, expr):
    pass 
  
  def visit_generic_expr(self, expr):
    for v in expr.children():
      self.visit_expr(v)

  _expr_method_names = {
    Var : 'visit_Var', 
    Const : 'visit_Const', 
    PrimCall : 'visit_PrimCall', 
    Attribute : 'visit_Attribute', 
    Index : 'visit_Index', 
    Tuple : 'visit_Tuple', 
    TupleProj : 'visit_TupleProj', 
    Slice : 'visit_Slice', 
    Struct : 'visit_Struct', 
    AllocArray : 'visit_AllocArray', 
    ArrayView : 'visit_ArrayView', 
    Array : 'visit_Array',
    Range : 'visit_Range',
    Ravel : 'visit_Ravel',   
    Transpose : 'visit_Transpose', 
    Shape : 'visit_Shape', 
    Strides : 'visit_Strides', 
    Alloc : 'visit_Alloc', 
    Cast : 'visit_Cast', 
    Call : 'visit_Call', 
    Select : 'visit_Select', 
    Map : 'visit_Map',
    OuterMap : 'visit_OuterMap',
    IndexMap : 'visit_IndexMap', 
    Reduce : 'visit_Reduce',
    IndexReduce : 'visit_IndexReduce',
    Scan : 'visit_Scan', 
    IndexScan : 'visit_IndexScan',
    Closure : 'visit_Closure', 
    ClosureElt : 'visit_ClosureElt', 
    UntypedFn : 'visit_UntypedFn',  
    TypedFn : 'visit_TypedFn', 
    TypeValue : 'visit_TypeValue', 
  }
  
  def visit_expr(self, expr):   
    c = expr.__class__
    if c is Var:  
      return self.visit_Var(expr)
    elif c is Const: 
      return self.visit_Const(expr)
    elif c is PrimCall:
      return self.visit_PrimCall(expr)
    elif c is Tuple:
      return self.visit_Tuple(expr)
    elif c is TupleProj:
      return self.visit_TupleProj(expr) 
    elif c is Index: 
      return self.visit_Index(expr)
    
    # try looking up the method in fast-path dictionary
    method_name = self._expr_method_names.get(c)
    if not method_name:
      # ...otherwise, just construct a new string following the visit_Expr pattern 
      method_name = "visit_" + c.__name__
    
    # if we found a method, call it 
    if method_name:
      method = getattr(self, method_name)
      
    if method:
      return method(expr)

    for child in expr.children():
      self.visit_expr(child)
      
  def visit_expr_list(self, exprs):
    return [self.visit_expr(expr) for expr in exprs]

  def visit_lhs_Var(self, lhs):
    self.visit_Var(lhs)

  def visit_lhs_Tuple(self, lhs):
    self.visit_Tuple(lhs)

  def visit_lhs_Index(self, lhs):
    self.visit_Index(lhs)

  def visit_lhs_Attribute(self, lhs):
    self.visit_Attribute(lhs)

  def visit_lhs(self, lhs):
    c = lhs.__class__
    if c is Var:
      return self.visit_lhs_Var(lhs)
    elif c is Tuple:
      return self.visit_lhs_Tuple(lhs)
    elif c is Index:
      return self.visit_lhs_Index(lhs)
    elif c is Attribute:
      return self.visit_lhs_Attribute(lhs)
    else:
      assert False, "LHS not implemented: %s" % (lhs,)

  def visit_block(self, stmts):
    for s in stmts:
      self.visit_stmt(s)

  def visit_Assign(self, stmt):
    self.visit_lhs(stmt.lhs)
    self.visit_expr(stmt.rhs)

  def visit_merge(self, phi_nodes):
    for (_, (l,r)) in phi_nodes.iteritems():
      self.visit_expr(l)
      self.visit_expr(r)

  def visit_merge_if(self, phi_nodes):
    self.visit_merge(phi_nodes)

  def visit_If(self, stmt):
    self.visit_expr(stmt.cond)
    self.visit_block(stmt.true)
    self.visit_block(stmt.false)
    self.visit_merge_if(stmt.merge)

  def visit_ExprStmt(self, stmt):
    self.visit_expr(stmt.value)

  def visit_Return(self, stmt):
    self.visit_expr(stmt.value)

  def visit_merge_loop_start(self, phi_nodes):
    pass

  def visit_merge_loop_repeat(self, phi_nodes):
    self.visit_merge(phi_nodes)

  def visit_While(self, stmt):
    self.visit_merge_loop_start(stmt.merge)
    self.visit_expr(stmt.cond)
    self.visit_block(stmt.body)
    self.visit_merge_loop_repeat(stmt.merge)
    
    
  def visit_ForLoop(self, stmt):
    self.visit_lhs(stmt.var)
    self.visit_expr(stmt.start)
    self.visit_merge_loop_start(stmt.merge)
    self.visit_expr(stmt.stop)
    self.visit_block(stmt.body)
    self.visit_expr(stmt.step)
    self.visit_merge_loop_repeat(stmt.merge)

  def visit_Comment(self, stmt):
    pass

  def visit_ParFor(self, expr):
    self.visit_expr(expr.fn)
    self.visit_expr(expr.bounds)
  
  _stmt_method_names = {
    Assign : 'visit_Assign', 
    Return : 'visit_Return', 
    While : 'visit_While', 
    ForLoop : 'visit_ForLoop', 
    If : 'visit_If', 
    ExprStmt : 'visit_ExprStmt', 
    ParFor : 'visit_ParFor', 
    Comment : 'visit_Comment',                    
  }
  
  def visit_stmt(self, stmt):
    c = stmt.__class__
    if c is Assign: 
      self.visit_Assign(stmt)
    else:
      getattr(self, self._stmt_method_names[c])(stmt)
    

  def visit_fn(self, fn):
    self.visit_block(fn.body)



import numpy as np
from parakeet.testing_helpers import expect, run_local_tests

def return_pair():
  return (-1.0, 200)

def test_return_pair():
  expect(return_pair, [], (-1.0, 200))

ints = (np.int32(1), np.int32(2))
mixed = (1.0, 200L)
nested = (ints, mixed)
nested2 = (nested, nested)

def all_tuples(f, unpack_args = True):
  """
  Given a function which should act as the identity, test it on multiple tuples
  """

  for t in [ints, mixed, nested2, nested2]:
    if unpack_args:
      expect(f, t, t)
    else:
      expect(f, [t], t)

def create_tuple(x,y):
  return (x,y)

def test_create_tuple():
  all_tuples(create_tuple)

#def tuple_arg((x,y)):
#  return (x,y)

#def test_tuple_arg():
#  all_tuples(tuple_arg, unpack_args = False)

def tuple_lhs(t):
  x,y = t
  return x,y

def test_tuple_lhs():
  all_tuples(tuple_lhs, unpack_args = False)

def tuple_lhs_sum(t):
  x, y, z = t
  return x + y + z

def test_tuple_lhs_sum():
  tuples = [(True, 0, 1.0), (1, True, 0)]
  for t in tuples:
    expect(tuple_lhs_sum, [t], sum(t))

def tuple_indexing(t):
  return (t[0], t[1])

def test_tuple_indexing():
  all_tuples(tuple_indexing, unpack_args = False)

def or_elts((b1,b2)):
  if b1 or b2:
    return 1
  else:
    return 0

def test_or_elts():
  expect(or_elts, [(True, True)], 1)
  expect(or_elts, [(True, False)], 1)
  expect(or_elts, [(False, True)], 1)
  expect(or_elts, [(False, False)], 0)

if __name__ == '__main__':
  run_local_tests()



import numpy as np

from parakeet.testing_helpers import expect, run_local_tests

def create_const(x):
  return [x,x,x,x]

def test_create_const():
  expect(create_const, [1],  np.array([1,1,1,1]))
  expect(create_const, [1.0], np.array([1.0, 1.0, 1.0, 1.0]))
  expect(create_const, [True], np.array([True, True, True, True]))


def test_nested_2d():
  expect(create_const, [np.array([1,2])] , np.array([[1,2],[1,2],[1,2],[1,2]]))
  expect(create_const, [np.array([1.0,2.0])] , np.array([[1.0,2.0],[1.0,2.0],[1.0,2.0],[1.0,2.0]]))
  expect(create_const, [np.array([True,False])] , 
         np.array([[True,False],[True,False],[True,False],[True,False]]))
  
if __name__ == '__main__':
  run_local_tests()



from parakeet.testing_helpers import expect, run_local_tests
import numpy as np 

global_int = 5 
global_tuple = (5,3.0)
global_array = np.array([1,2,3])

def use_globals(x):
  return x + global_int + global_array + global_tuple[1]

def test_use_globals():
  expect(use_globals, [2], use_globals(2))

def use_locals(x):
  local_int = global_int
  local_array = global_array 
  local_tuple = global_tuple
  local_int2 = 3
  local_tuple2 = (5.0,9.0)
  local_array2 = np.array([4,5,6])
  def nested(y):
    return x + y + local_int + local_array + local_tuple[1] + local_int2 + local_tuple2[1] + local_array2
  return nested(x+1)

def test_use_locals():
  expect(use_locals, [True], use_locals(True))

if __name__ == "__main__":
  run_local_tests()



from itertools import izip 

from .. import config

from .. syntax import TypedFn
from clone_function import CloneFunction
from recursive_apply import RecursiveApply
from transform import Transform 

name_stack = []
def apply_transforms(fn, transforms, 
                       cleanup = [], 
                       phase_name = None, 
                       transform_history = None):
  if len(transforms) == 0:
    return fn 
  if phase_name: name_stack.append("{" + phase_name + " :: " + fn.name +  "}")
  for T in transforms:
    t = T() if type(T) == type else T
    
    if isinstance(t, Transform):
      name_stack.append(str(t))
      if config.print_transform_names:
        print "-- %s" % ("->".join(name_stack),)
      
    elif isinstance(t, Phase) and t.should_skip(fn) and not t.depends_on:
      continue 

    fn = t.apply(fn)

    assert fn is not None, "%s transformed fn into None" % T

    if isinstance(t, Transform): name_stack.pop()

    if len(cleanup) > 0:
      fn = apply_transforms(fn, cleanup, [], phase_name = "cleanup")
    
    if transform_history is not None:
      transform_history.add(T)
  
  if phase_name: name_stack.pop()
  return fn

name_stack = []
class Phase(object):
  def __init__(self,
               transforms,
               depends_on = [],
               copy = False,
               cleanup = [],
               config_param = None,
               run_if = None,
               rename = False,
               post_apply = None,
               memoize = True,
               name = None, 
               recursive = True):
    self.cache = {}
    if not isinstance(transforms, (tuple, list)):
      transforms = [transforms]
    self.transforms = transforms

    if not depends_on:
      depends_on = []
    if not isinstance(depends_on, (tuple, list)):
      depends_on = [depends_on]
    self.depends_on = depends_on

    self.copy = copy

    if not cleanup:
      cleanup = []
    if not isinstance(cleanup, (list,tuple)):
      cleanup = [cleanup]
    self.cleanup = cleanup

    self.config_param = config_param
    self.run_if = run_if
    self.rename = rename
    self.post_apply = post_apply
    self.memoize = memoize
    self.name = name
    self.recursive = recursive 
    self._hash = hash(str(self))

  def __str__(self):
    if self.name:
      return self.name
    else:
      names = []
      for t in self.transforms:
        if type(t) == type:
          names.append(t.__name__)
        else:
          names.append(str(t))
      if len(names) == 1:
        return names[0]
      else:
        return "Phase(%s)" % (",".join(names))

  def __repr__(self):
    return str(self)

  def __eq__(self, other):
    if self is other:
      return True
    
    else:
      return (self.__class__ is other.__class__) and\
             (self.name == other.name) and \
             (self.copy == other.copy) and \
             (self.memoize == other.memoize) and \
             (len(self.transforms) == len(other.transforms)) and \
             (all(t1 == t2 for t1, t2 in izip(self.transforms, other.transforms))) 

  def __hash__(self):
    return self._hash 
  
  def __call__(self, fn, run_dependencies = True):
    return self.apply(fn, run_dependencies)

  def should_skip(self, fn):
    if self.config_param is not None and getattr(config, self.config_param) == False:
      return True
   
    if self.memoize and  self in fn.transform_history:
      return True
    
    if self.run_if:
      return not self.run_if(fn)
    
    return False 
    
  def is_cached(self, fn):
    return fn.cache_key in self.cache 
  
  def needs_cleanup(self, fn):
    return not (self.should_skip(fn) or self.is_cached(fn)) 
    
  def apply(self, fn, run_dependencies = True):
    
      
    if self.memoize:
      original_key = fn.cache_key
      if fn.created_by is self:
        return fn 
      if original_key in self.cache:
        return self.cache[original_key] 
    
    if self.depends_on and run_dependencies:
      fn = apply_transforms(fn, self.depends_on)
    
    if self.copy:
      fn = CloneFunction(parent_transform = self, rename = self.rename).apply(fn)
      if fn.cache_key  in self.cache:
        print "Warning: Typed function %s (key = %s) already registered, encountered while cloning before %s" % \
        (fn.name, fn.cache_key, self)
    
    if self.recursive:
      fn = RecursiveApply(self).apply(fn)
      
    if not self.should_skip(fn):
      fn = apply_transforms(fn, self.transforms, 
                           cleanup = self.cleanup, 
                           phase_name = str(self), 
                           transform_history = fn.transform_history)
    
      if self.post_apply:
        new_fn = self.post_apply(fn)
        if new_fn.__class__ is TypedFn:
          fn = new_fn

    fn.transform_history.add(self)
      
    if self.memoize:
      self.cache[original_key] = fn
      self.cache[fn.cache_key] = fn
    return fn



from parakeet.testing_helpers import expect, run_local_tests

def test_float_to_int():
  expect(int, [1.2], 1)

def test_int_to_float():
  expect(float, [1], 1.0)
  
def test_bool_to_float():
  expect(float, [True], 1.0)

def test_float_to_long():
  expect(long, [-1.0], -1)
  
if __name__ == '__main__':
  run_local_tests()




from dsltools import ScopedSet 

from .. analysis.collect_vars import collect_var_names, collect_binding_names
from .. analysis.escape_analysis import may_alias
from .. analysis.syntax_visitor import SyntaxVisitor

from .. ndtypes import ImmutableT, ArrayT, PtrT 
from .. syntax import Var, Assign, Return, While, If, Index, Alloc, AllocArray  
from .. syntax import Array, ArrayView, Slice, Struct 

from transform import Transform

class Find_LICM_Candidates(SyntaxVisitor):
  def __init__(self):
    SyntaxVisitor.__init__(self)
    self.mutable_types = None
    self.volatile_vars = ScopedSet()
    self.depends_on = {}
    self.safe_to_move = set([])
    self.curr_block_id = None
    self.block_contains_return = set([])
    self.may_alias = None

  def visit_fn(self, fn):
    self.volatile_vars.push(fn.arg_names)
    self.may_alias = may_alias(fn)
    SyntaxVisitor.visit_fn(self, fn)
    return self.safe_to_move

  def mark_safe_assignments(self, block, volatile_set):
    for stmt in block:
      klass = stmt.__class__
      if klass is Assign and \
         stmt.lhs.__class__ is Var:
        name = stmt.lhs.name
        dependencies = self.depends_on.get(name, set([]))
        volatile = name in volatile_set or \
                   any(d in volatile_set for d in dependencies)
        if not volatile:
          self.safe_to_move.add(name)
      # just in case there are Returns in nested control flow
      # we should probably avoid changing the performance characteristics
      # by pulling out statements which will never run
      elif klass is If:
        if id(stmt.true) in self.block_contains_return or \
           id(stmt.false) in self.block_contains_return:
          break
      elif klass is While:
        if id(stmt.body) in self.block_contains_return:
          break
      elif klass is Return:
        break

  def mark_curr_block_returns(self):
    self.block_contains_return.add(self.curr_block_id)

  def does_block_return(self, block):
    return id(block) in self.block_contains_return

  def visit_Return(self, stmt):
    self.mark_curr_block_returns()

  def visit_block(self, stmts):
    self.curr_block_id = id(stmts)
    SyntaxVisitor.visit_block(self, stmts)

  def visit_merge(self, merge, both_branches = True):
    pass

  def visit_ForLoop(self, stmt):
    self.volatile_vars.push(stmt.merge.keys())
    self.volatile_vars.add(stmt.var.name)
    SyntaxVisitor.visit_ForLoop(self, stmt)
    if self.does_block_return(stmt.body):
      self.block_contains_return()
    volatile_in_scope = self.volatile_vars.pop()
    self.mark_safe_assignments(stmt.body, volatile_in_scope)

  def visit_While(self, stmt):
    self.volatile_vars.push(stmt.merge.keys())
    SyntaxVisitor.visit_While(self, stmt)
    if self.does_block_return(stmt.body):
      self.block_contains_return()
    volatile_in_scope = self.volatile_vars.pop()
    self.mark_safe_assignments(stmt.body, volatile_in_scope)

  def visit_Var(self, expr):
    self.volatile_vars.add(expr.name)

  def visit_If(self, stmt):
    self.volatile_vars.push(stmt.merge.keys())
    self.visit_expr(stmt.cond)
    SyntaxVisitor.visit_If(self, stmt) 
    if self.does_block_return(stmt.true) or self.does_block_return(stmt.false):
      self.mark_curr_block_returns()
    volatile_in_scope = self.volatile_vars.pop()
    self.mark_safe_assignments(stmt.true, volatile_in_scope)
    self.mark_safe_assignments(stmt.false, volatile_in_scope)

  def is_mutable_alloc(self, expr):
    return isinstance(expr.type, (PtrT, ArrayT)) 
    #c = expr.__class__
    #return c in (Alloc, AllocArray, )
    #return c is Alloc or \
    #       c is AllocArray or \
    #       c is Array or \
    #       c is ArrayView or \
    #       c is 
    #       (c is Struct and not isinstance(expr.type, ImmutableT))

  def visit_Assign(self, stmt):
     
    lhs_names = collect_binding_names(stmt.lhs)
    rhs_names = collect_var_names(stmt.rhs)
   
    for x in lhs_names:
      dependencies = self.depends_on.get(x, set([]))
      dependencies.update(rhs_names)
      self.depends_on[x] = dependencies

    if any(x in self.volatile_vars for x in rhs_names):
      self.volatile_vars.update(lhs_names)
    elif self.is_mutable_alloc(stmt.rhs):
      if len(lhs_names) == 1 and \
         len(self.may_alias.get(lhs_names[0], [])) <= 1:
        pass
      else:
        self.volatile_vars.update(lhs_names)
    # mark any array writes as volatile 
    if stmt.lhs.__class__ is Index:
      assert stmt.lhs.value.__class__ is Var, \
        "Expected LHS array to be variable but instead got %s" % stmt  
      self.volatile_vars.add(stmt.lhs.value.name)
      #print 
      #print "STMT", stmt
      #print "lhs names", lhs_names 
      #print "rhs names", rhs_names 
      #print "volatile vars", self.volatile_vars
      #print 
      

class LoopInvariantCodeMotion(Transform):
  def __init__(self):
    Transform.__init__(self)

  def pre_apply(self, fn):
    self.analysis = Find_LICM_Candidates()
    self.safe_to_move = self.analysis.visit_fn(fn)
    self.binding_depth = {}
    self.mark_binding_depths(fn.arg_names, 0)

  def mark_binding_depths(self, names, depth_offset = 0):
    curr_depth = len(self.blocks._blocks) - 1 + depth_offset
    for name in names:
      self.binding_depth[name] = curr_depth

  def transform_ForLoop(self, stmt):
    self.mark_binding_depths(stmt.merge.iterkeys(), 1)
    self.mark_binding_depths([stmt.var.name], 1)
    return Transform.transform_ForLoop(self, stmt)

  def transform_While(self, stmt):
    self.mark_binding_depths(stmt.merge.iterkeys(), 1)
    return Transform.transform_While(self, stmt)

  def transform_If(self, stmt):
    self.mark_binding_depths(stmt.merge.iterkeys(), 1)
    return Transform.transform_If(self, stmt)

  def transform_Assign(self, stmt):
    # TODO: Allow possibility of indexing into variables that are read-only
    if stmt.lhs.__class__ is Var and stmt.rhs.__class__ is not Index:
      name = stmt.lhs.name
      if name in self.safe_to_move:
        deps = self.analysis.depends_on[name]

        if all(d in self.binding_depth for d in deps):
          if len(deps) > 0:
            target_level = max(self.binding_depth[d] for d in deps)
          else:
            target_level = 0
 
          if target_level >= 0 and target_level < self.blocks.depth():
            self.blocks._blocks[target_level].append(stmt)
            self.binding_depth[name] = target_level
            return None
    self.mark_binding_depths(collect_binding_names(stmt.lhs))
    return stmt




from parakeet import jit
import numpy as np 

from timer import timer 


def compare_perf(fn, args, numba= True, cpython = True, 
                 extra = {}, 
                 backends = ('c', 'openmp', 'cuda'), 
                 suppress_output = False,
                 propagate_exceptions = False):

  
  parakeet_fn = jit(fn)
  name = fn.__name__
  parakeet_result = None
  numba_result = None 
  cpython_result = None 
  kwargs = {'suppress_stdout': suppress_output, 
            'suppress_stderr':suppress_output,
            'propagate_exceptions' : propagate_exceptions
           }
  backend = None 
  for backend in backends:
    with timer('Parakeet (backend = %s) #1 -- %s' % (backend, name), **kwargs):
      parakeet_result = parakeet_fn(*args, _backend = backend)

    with timer('Parakeet (backend = %s) #2 -- %s' % (backend, name), **kwargs):
      parakeet_result = parakeet_fn(*args, _backend = backend)

  if numba:
    from numba import autojit, config 
    numba_fn = autojit(fn)
    with timer('Numba #1 -- %s' % name, **kwargs):
      numba_result = numba_fn(*args)

    with timer('Numba #2 -- %s' % name, **kwargs):
      numba_result = numba_fn(*args)

  if cpython:
    with timer('Python -- %s' % name, **kwargs):
      cpython_result = fn(*args)
  

  rtol = 0.0001
  if backend in ('cuda', 'gpu'):
    atol = 0.001
  else:
    atol = 0.00001
  if parakeet_result is not None and cpython_result is not None:
      diffs = np.abs(cpython_result - parakeet_result)
      assert np.allclose(cpython_result, parakeet_result, atol = atol, rtol = rtol), \
        "Max elt difference between Parakeet and CPython = %s (median = %s, min = %s)" % \
        (np.max(diffs), np.median(diffs), np.min(diffs))
  
  if numba_result is not None and cpython_result is not None:
      diffs = np.abs(cpython_result - numba_result)
      assert np.allclose(cpython_result, numba_result, atol = atol, rtol = rtol), \
        "Max elt difference between Numba and CPython = %s (median = %s, min = %s)" % \
        (np.max(diffs), np.median(diffs), np.min(diffs))


  
  for name, impl in extra.iteritems():
    with timer("%s #1" % name, **kwargs):
      impl(*args)
    with timer("%s #2" % name, **kwargs):
      extra_result = impl(*args)
    if parakeet_result is not None:
      diffs = np.abs(parakeet_result - extra_result)
      assert np.allclose(parakeet_result, extra_result, atol = atol, rtol = rtol), \
        "Max elt difference between Parakeet and %s = %s (median = %s, min = %s)" % \
        (name, np.max(diffs), np.median(diffs), np.min(diffs))




from .. ndtypes import TypeValueT 
from expr import Expr 


class TypeValue(Expr):
  """
  Value materialization of a type 
  """
  def __init__(self, type_value, type = None, source_info = None):
    self.type_value = type_value
     
    if type is None: type = TypeValueT(self.type_value)
    assert isinstance(type, TypeValueT)
    assert type.type is not None 
    
    self.type = type 
     
    


from run_function import specialize

def typed_repr(python_fn, args, optimize = True):
  typed_fn, _ = specialize(python_fn, args)
  if optimize:
    from ..transforms import pipeline
    return pipeline.high_level_optimizations.apply(typed_fn)
  else:
    return typed_fn 
  
  



keywords = set([
  "auto", 
  "break",
  "case",
  "char",
  "const",
  "continue", 
  "default",
  "do",
  "double",
  "else",
  "enum",
  "extern", 
  "float",
  "for",
  "goto",
  "if", 
  "int",
  "long",
  "register",
  "return"
  "short"
  "signed"
  "sizeof"
  "static"
  "struct",
  "switch"
  "typedef", 
  "union", 
  "unsigned",
  "void", 
  "volatile",
  "while"])

macro_names = set(["assert"])
util_names = set(["printf", "malloc", "free", ])

import math 

base_math_names = [name for name in dir(math) if not name.startswith("__")]

float32_math_names = [name+'f' for name in base_math_names]
long_math_names = [name+"l" for name in base_math_names]
extra_math_names = [
  "atof", "cbrt", "cbrtf", "cbrtl",
  "div", 
  "exp2", "exp2f", "exp2l", 
  "j0", "j1", "jn",
  "log2", "log2f", "log2l", 
  "logb", "logbf", "logbl", 
  "lrintf", "lrintl", 
  "lround", "lroundf", "lroundl", 
  "remainder", "remainderf", "remainderl", 
  "remquo", "remquof", "remquol", 
  "rint", "rintf", "rintl", 
  "round", "roundf", "roundl",
  "scalb", "scalbln", "scalblnf", "scalblnl", "scalbn", "scalbnf", "scalbnl",  
  "signgam", 
  "tgamma", "tgammaf", "tgammal"  
  "y0", "y1", "yn",              
]
math_names = set(base_math_names+float32_math_names+long_math_names+extra_math_names)
all_reserved_names = keywords.union(macro_names).union(util_names).union(math_names)
 
def is_reserved(name):
  return name in all_reserved_names or name.startswith("Py") 



_type_mapping = {}
_typeof_functions = {}

# python type -> (python value -> internal value) 
_from_python_fns = {}
# class of ndtype -> (internal value -> python value)
_to_python_fns = {}
def register(python_types, 
              parakeet_type, 
              typeof = None, 
              to_python = None, 
              from_python = None):
  """
  Map each python type to either a parakeet type or a function that returns a
  parakeet type
  """

  if typeof is None:
    typeof = lambda _: parakeet_type

  if not isinstance(python_types, (list, tuple)):
    python_types = [python_types]

  for python_type in python_types:
    _typeof_functions[python_type] = typeof
    _type_mapping[python_type] = parakeet_type
    
  if to_python is not None:
    _to_python_fns[parakeet_type] = to_python
           
  if from_python is not None:
    for python_type in python_types:
      _from_python_fns[python_type] = from_python     
     
  
def equiv_type(python_type):
  assert python_type in _type_mapping, \
      "No type mapping found for %s" % python_type
  return _type_mapping[python_type]

class InvalidType(Exception):
  def __init__(self, python_value, python_type):
    self.python_value = python_value
    self.python_type = python_type 
    
  def __str__(self):
    return "Don't know how to convert value %s : %s" % (self.python_value, self.python_type)

def typeof(python_value):
  python_type = type(python_value)
  try:
    parakeet_type_lookup = _typeof_functions[python_type]
  except:
    raise InvalidType(python_value, python_type)
  return parakeet_type_lookup(python_value)
  
def from_python(python_value):
  """
  Look up the ctypes representation of the corresponding parakeet type and call
  the converter with the ctypes class and python value
  """
  python_type = type(python_value)
  if python_type in _from_python_fns:
    return _from_python_fns[python_type](python_value)
  parakeet_type = typeof(python_value)
  return parakeet_type.from_python(python_value)

def to_python(internal_value, parakeet_type):
  if parakeet_type.__class__ in _to_python_fns:
    return _to_python_fns[parakeet_type.__class__](internal_value)
  return parakeet_type.to_python(internal_value)
  



from parakeet import jit 
import numpy as np 


def dist(x,y):
  return ((x-y)**2).sum()

@jit 
def kmeans(X, k, niters = 10):
  C = X[:k, :]
  for _ in xrange(niters):
    A = np.array([np.argmin([dist(x,c) for c in C]) for x in X])
    C = np.array([np.mean(X[A == i, :], axis = 0) for i in xrange(k)])
  return C


n, d = 10**3, 100
X = np.random.randn(n,d)
k = 5

C = kmeans(X, k, niters=20) 
print "Clusters:", C





######################################
#        OPTIMIZER OPTIONS           #
######################################

# run LLVM optimization passes
llvm_optimize = True

# number of times to run optimizations
llvm_num_passes = 4

# run verifier over generated LLVM code?
llvm_verify = True


######################################
#         PRINTING OPTIONS           #
######################################

# print generated assembly of compiled functions
print_x86 = False

# show LLVM bytecode before optimization passes
print_unoptimized_llvm = False

# show LLVM bytecode after optimizations
# print_optimized_llvm = False 




from ..ndtypes import Int32 
from ..syntax import SourceExpr, SourceStmt

class dim3(object):
  def __init__(self, x, y, z):
    self.x = x 
    self.y = y 
    self.z = z 
  
  def __iter__(self):
    yield self.x 
    yield self.y 
    yield self.z
    
  def __getitem__(self, idx):
    if idx == 0:
      return self.x 
    elif idx == 1:
      return self.y 
    else:
      assert idx == 2, "Unexpected index %s to %s" % (idx, self)
      return self.z
    
  def __str__(self):
    return "dim3(x = %s, y = %s, z = %s)" % (self.x, self.y, self.z) 
 
# pasted literally into the C source 
blockIdx_x = SourceExpr("blockIdx.x", type=Int32)
blockIdx_y = SourceExpr("blockIdx.y", type=Int32)
blockIdx_z = SourceExpr("blockIdx.z", type=Int32)
blockIdx = dim3(blockIdx_x, blockIdx_y, blockIdx_z)

threadIdx_x = SourceExpr("threadIdx.x", type=Int32)
threadIdx_y = SourceExpr("threadIdx.y", type=Int32)
threadIdx_z = SourceExpr("threadIdx.z", type=Int32)
threadIdx = dim3(threadIdx_x, threadIdx_y, threadIdx_z)

blockDim_x = SourceExpr("blockDim.x", type = Int32)
blockDim_y = SourceExpr("blockDim.y", type = Int32)
blockDim_z = SourceExpr("blockDim.z", type = Int32)
blockDim = dim3(blockDim_x, blockDim_y, blockDim_z)

gridDim_x = SourceExpr("gridDim.x", type = Int32)
gridDim_y = SourceExpr("gridDim.y", type = Int32)
gridDim_z = SourceExpr("gridDim.z", type = Int32)
gridDim = dim3(gridDim_x, gridDim_y, gridDim_z)

warpSize = SourceExpr("wrapSize", type=Int32)
__syncthreads = SourceStmt("__syncthreads;")




from parakeet import jit  

@jit 
def avg(x,y,z):
  """
  Return the average of three scalars or arrays
  (gets optimized into single traversal)
  """
  return (x + y + z) / 3.0

import numpy as np 

N = 20
x = np.random.randn(N)
y = np.random.randn(N)
z = np.random.randn(N)

assert np.allclose(avg(x,y,z), (x+y+z)/3.0)






import numpy as np
import parakeet 

from parakeet.testing_helpers import expect, run_local_tests 


def test_empty():
  s = (20, 20, 3)
  x = np.empty(s)
  assert x.shape == s
  
def test_empty_int():
  s = (2,2,2,2)
  x = np.empty(s, dtype=np.uint8)
  assert x.shape == s
  assert x.dtype == np.uint8


if __name__ == '__main__':
  run_local_tests()
